#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
"""Contains classes for conversion between ADS runtime and OCI Data Science Job implementation.
This module is for ADS developers only.
In this module, a payload is defined as a dictionary for initializing a DSCJob object.
The DSCJob can be initialized with the same arguments for initializing oci.data_science.models.Job,
    plus an "artifact" argument for job artifact.
The payload also contain infrastructure information.
The conversion from a runtime to a payload is called translate in this module.
The conversion from a DSCJob to a runtime is called extract in this module.
"""
from __future__ import annotations

import json
import os
import shlex
from typing import Optional
from urllib import parse
from ads.common.utils import extract_region
from ads.jobs.builders.runtimes.base import Runtime
from ads.jobs.builders.runtimes.python_runtime import (
    CondaRuntime,
    ScriptRuntime,
    PythonRuntime,
    NotebookRuntime,
    GitPythonRuntime,
)
from ads.jobs.builders.runtimes.container_runtime import ContainerRuntime
from ads.jobs.builders.runtimes.artifact import (
    ScriptArtifact,
    NotebookArtifact,
    PythonArtifact,
    GitPythonArtifact,
)
from ads.jobs.builders.infrastructure.utils import get_value


class IncompatibleRuntime(Exception):
    """Represents an exception when runtime is not compatible with the OCI data science job configuration.
    This exception is designed to be raised during the extraction of a runtime from OCI data science job.
    The data science job does not explicitly contain information of the type of the ADS runtime.
    Each runtime handler should determine if the configuration of the job can be converted to the runtime.
    This exception should be raised during the extract() call if the configuration cannot be converted.
    The RuntimeManager uses this exception to determine if the conversion is successful.
    """


class RuntimeHandler:
    """Base class for Runtime Handler.

    Each runtime handler should define the RUNTIME_CLASS to be the runtime it can handle.

    Each runtime handler is initialized with a DataScienceJob instance.
    This instance is a reference and the modification will be exposed to the users.

    Each runtime handler expose two methods: translate() and extract().
    In this class, translate or extract signals the direction of conversion.
    All method starts with "translate" handles the conversion from ADS runtime to OCI API payload.
    All method starts with "extract" handles the conversion from OCI data science Job to ADS runtime.
    This base class defines the default handling for translate() and extract().
    Each sub-class can override the two methods to provide additional handling.
    Alternatively, a sub-class can also override a sub-method, which is called by the translate() or extract() method.
    For example, _translate_env() handles the conversion of environment variables from ADS runtime to OCI API payload.

    See the individual methods for more details.
    """

    # Defines the class of the runtime to be handled.
    RUNTIME_CLASS = Runtime

    def __init__(self, data_science_job) -> None:
        """Initialize the runtime handler.

        Parameters
        ----------
        data_science_job : DataScienceJob
            An instance of the DataScienceJob to be created or extracted from.
        """
        self.data_science_job = data_science_job

    def translate(self, runtime: Runtime) -> dict:
        """Translates the runtime into a JSON payload for OCI API.
        This method calls the following sub-methods:
        * _translate_artifact()
        * _translate_config()
          * _translate_env()
        A sub-class can modify one of more of these methods.

        Parameters
        ----------
        runtime : Runtime
            An instance of the runtime to be converted to a JSON payload.

        Returns
        -------
        dict
            JSON payload for defining a Data Science Job with OCI API
        """
        payload = {}
        payload["artifact"] = self._translate_artifact(runtime)
        payload["job_configuration_details"] = self._translate_config(runtime)
        if runtime.freeform_tags:
            payload["freeform_tags"] = runtime.freeform_tags
        self.data_science_job.runtime = runtime
        return payload

    def _translate_artifact(self, runtime: Runtime):
        """Translate the runtime artifact.
        OCI data science requires an artifact file to be uploaded before the job is created.
        For Python runtime, the artifact is the script for running the job.
        For container runtime, the artifact is not actually used.
        For notebook runtime, ADS needs to convert the artifact to Python script before uploading.

        Parameters
        ----------
        runtime : Runtime
            An instance of the runtime

        Returns
        -------
        str or Artifact
            The artifact that is ready to be used by DSCJob.
            This can either be a string storing the path the artifact file,
            or an instance of Artifact class, which contains logic for additional processing.
        """
        return None

    def _translate_env(self, runtime: Runtime) -> dict:
        """Translate the environment variable.

        OCI Data Science job uses environment variables for various settings.
        These settings are properties in ADS runtime.
        This method is designed to handle the conversion of the ADS runtime properties to environment variables.
        By default, no conversion is made in this method.
        Sub-class should override this method to add conversion logic.

        Parameters
        ----------
        runtime : Runtime
            An instance of the runtime

        Returns
        -------
        dict
            A dictionary storing the environment variable for OCI data science job.
        """
        return runtime.envs

    def _translate_config(self, runtime: Runtime) -> dict:
        """Prepares the job configuration from runtime specifications.

        Parameters
        ----------
        runtime : Runtime
            An instance of the runtime

        Returns
        -------
        dict
            A dictionary for OCI data science job configuration.
            The dictionary may have the following keys:
                "jobType"
                "commandLineArguments"
                "environmentVariables"
                "maximumRuntimeInMinutes"
            The configurations will be used to initialize the DSCJob instance.
            The DSCJob class can handle keys in either camel or snake format.
        """
        job_configuration_details = {
            "jobType": self.data_science_job.job_type,
        }
        if runtime.maximum_runtime_in_minutes:
            job_configuration_details[
                "maximum_runtime_in_minutes"
            ] = runtime.maximum_runtime_in_minutes
        job_configuration_details["environment_variables"] = self._translate_env(
            runtime
        )
        if runtime.args:
            # shlex.join() is not available until python 3.8
            job_configuration_details["command_line_arguments"] = " ".join(
                shlex.quote(arg) for arg in runtime.args
            )
        return job_configuration_details

    @staticmethod
    def _translate_specs(
        runtime: Runtime, spec_mappings: dict, delimiter: Optional[str] = None
    ) -> dict:
        """Converts runtime properties to OCI data science job environment variables based on a mapping.

        Parameters
        ----------
        runtime : Runtime
            The runtime containing the properties to be converted.
        spec_mappings : dict
            Mapping from runtime properties to environment variables.
            Each key is a specification key (property name) of a runtime
            Each value is the corresponding name of the environment variable in OCI data science job.
        delimiter : str, Optional
            Environment variables must be strings.
            For list or tuple, specify the delimiter for joining the values into a string.

        Returns
        -------
        dict
            A dictionary contianing environment variables for OCI data science job.
        """
        envs = {}
        for spec_key, dsc_key in spec_mappings.items():
            val = runtime.get_spec(spec_key)
            if val:
                if delimiter and isinstance(val, list) or isinstance(val, tuple):
                    val = delimiter.join(val)
                envs[dsc_key] = val
        return envs

    @staticmethod
    def _extract_specs(envs: dict, spec_mappings: dict) -> dict:
        """Converts the environment variables in OCI data science job to runtime properties.

        Parameters
        ----------
        envs : dict
            A dictionary contianing environment variables from OCI data science job.
        spec_mappings : dict
            Mapping from runtime properties to environment variables.
            This mapping is the same as the one in _translate_spec().

        This method does not convert strings into list or tuple as there is no way to identify them.

        Returns
        -------
        dict
            A dictionary for specifying the runtime.
        """
        spec = {}
        for spec_key, dsc_key in spec_mappings.items():
            val = envs.pop(dsc_key, None)
            if val:
                spec[spec_key] = val
        return spec

    @staticmethod
    def _format_env_var(runtime_spec: dict) -> dict:
        """Formats the environment variables in runtime specification (as dict) from a dictionary to list.
        The list of environment variables uses the same format as environment variables in Kubernetes.

        Parameters
        ----------
        runtime_spec : dict
            Runtime specification in a dictionary.
            This is the dictionary that can be used to initialize a runtime instance.
            Except that environment variables are stored in a dict instead of list.

        Returns
        -------
        dict
            Runtime specification with environment variables stored in a list.
        """
        env_var = runtime_spec.pop(Runtime.CONST_ENV_VAR, None)
        if env_var and isinstance(env_var, dict):
            runtime_spec[Runtime.CONST_ENV_VAR] = [
                {"name": k, "value": v} for k, v in env_var.items()
            ]
        return runtime_spec

    def extract(self, dsc_job):
        """Extract the runtime from an OCI data science job object.
        This method calls the following sub-methods:
        * _extract_tags()
        * _extract_args()
        * _extract_envs()
        * _extract_artifact()
        * _extract_runtime_minutes()
        Each of these method returns a dict for specifying the runtime.
        The dictionaries are combined before initalizing the runtime.
        A sub-class can modify one of more of these methods.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        Runtime:
            The runtime extracted from the data science job.
        """
        runtime_spec = {}
        extractions = [
            self._extract_tags,
            self._extract_args,
            self._extract_envs,
            self._extract_artifact,
            self._extract_runtime_minutes,
        ]
        for extraction in extractions:
            runtime_spec.update(extraction(dsc_job))
        return self.RUNTIME_CLASS(self._format_env_var(runtime_spec))

    def _extract_args(self, dsc_job) -> dict:
        """Extracts the command line arguments from data science job.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        args_string = get_value(
            dsc_job, "job_configuration_details.command_line_arguments"
        )
        if args_string:
            return {Runtime.CONST_ARGS: shlex.split(args_string)}
        return {}

    def _extract_envs(self, dsc_job):
        """Extract the environment variables from data science job.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        envs = get_value(dsc_job, "job_configuration_details.environment_variables")
        if envs:
            return {Runtime.CONST_ENV_VAR: envs}
        return {}

    def _extract_tags(self, dsc_job):
        """Extract the freeform tags from data science job.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        value = get_value(dsc_job, "freeform_tags")
        if value:
            return {Runtime.CONST_TAG: value}
        return {}

    def _extract_artifact(self, dsc_job):
        """Extract the job artifact from data science job.

        This is the base method which does not extract the job artifact.
        Sub-class should implement the extraction if needed.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        return {}

    def _extract_runtime_minutes(self, dsc_job):
        """Extract the maximum runtime in minutes from data science job.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        maximum_runtime_in_minutes = get_value(
            dsc_job, "job_configuration_details.maximum_runtime_in_minutes"
        )
        if maximum_runtime_in_minutes:
            return {
                Runtime.CONST_MAXIMUM_RUNTIME_IN_MINUTES: maximum_runtime_in_minutes
            }
        return {}


class CondaRuntimeHandler(RuntimeHandler):
    """Runtime Handler for CondaRuntime"""

    RUNTIME_CLASS = CondaRuntime
    CONST_CONDA_TYPE = "CONDA_ENV_TYPE"
    CONST_CONDA_SLUG = "CONDA_ENV_SLUG"
    CONST_CONDA_OBJ_NAME = "CONDA_ENV_OBJECT_NAME"
    CONST_CONDA_REGION = "CONDA_ENV_REGION"
    CONST_CONDA_NAMESPACE = "CONDA_ENV_NAMESPACE"
    CONST_CONDA_BUCKET = "CONDA_ENV_BUCKET"

    def __get_auth_region(self) -> str:
        return extract_region(self.data_science_job.dsc_job.auth)

    def _translate_env(self, runtime: CondaRuntime) -> dict:
        """Translate the environment variable.

        Parameters
        ----------
        runtime : CondaRuntime
            An instance of CondaRuntime

        Returns
        -------
        dict
            A dictionary containing environment variables for OCI data science job.
        """
        envs = super()._translate_env(runtime)
        if runtime.conda:
            envs[self.CONST_CONDA_TYPE] = runtime.conda.get(
                CondaRuntime.CONST_CONDA_TYPE
            )
            if (
                runtime.conda.get(CondaRuntime.CONST_CONDA_TYPE)
                == CondaRuntime.CONST_CONDA_TYPE_SERVICE
            ):
                envs.update(
                    {
                        self.CONST_CONDA_SLUG: runtime.conda.get(
                            CondaRuntime.CONST_CONDA_SLUG
                        ),
                    }
                )
            elif (
                runtime.conda.get(CondaRuntime.CONST_CONDA_TYPE)
                == CondaRuntime.CONST_CONDA_TYPE_CUSTOM
            ):
                uri = runtime.conda.get(CondaRuntime.CONST_CONDA_URI)
                p = parse.urlparse(uri)
                if not (p.username and p.hostname and p.path):
                    raise ValueError(
                        f"Invalid URI for custom conda pack: {uri}. "
                        "A valid URI should have the format: oci://your_bucket@namespace/object_name"
                    )
                region = runtime.conda.get(CondaRuntime.CONST_CONDA_REGION)
                if not region:
                    region = self.__get_auth_region()
                if not region:
                    raise AttributeError(
                        "Unable to determine the region for the custom conda pack. "
                        "Specify the region using with_custom_conda(uri, region)."
                    )
                envs.update(
                    {
                        self.CONST_CONDA_NAMESPACE: p.hostname,
                        self.CONST_CONDA_BUCKET: p.username,
                        self.CONST_CONDA_OBJ_NAME: p.path.lstrip("/"),
                        self.CONST_CONDA_REGION: region,
                    }
                )
        return envs

    def _extract_envs(self, dsc_job) -> dict:
        """Extract the environment variables from data science job.
        CondaRuntime contains environment variables for specifying conda environment.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        spec = super()._extract_envs(dsc_job)
        envs = spec.pop(CondaRuntime.CONST_ENV_VAR, {})
        conda_spec = self.__extract_conda_env(envs)
        if conda_spec:
            spec[CondaRuntime.CONST_CONDA] = conda_spec
        if envs:
            spec[CondaRuntime.CONST_ENV_VAR] = envs
        return spec

    @staticmethod
    def __extract_conda_env(envs: dict) -> Optional[dict]:
        """Extracts conda pack specification from environment variables

        Parameters
        ----------
        envs : dict
            Environment variables.

        Returns
        -------
        Optional[dict]
            Conda pack runtime specification.
        """
        if not envs:
            return None
        if (
            CondaRuntimeHandler.CONST_CONDA_TYPE in envs
            and CondaRuntimeHandler.CONST_CONDA_SLUG in envs
        ):
            return {
                CondaRuntime.CONST_CONDA_TYPE: envs.pop(
                    CondaRuntimeHandler.CONST_CONDA_TYPE
                ),
                CondaRuntime.CONST_CONDA_SLUG: envs.pop(
                    CondaRuntimeHandler.CONST_CONDA_SLUG
                ),
            }
        if (
            envs.get(CondaRuntimeHandler.CONST_CONDA_TYPE)
            == CondaRuntime.CONST_CONDA_TYPE_CUSTOM
            and CondaRuntimeHandler.CONST_CONDA_BUCKET in envs
            and CondaRuntimeHandler.CONST_CONDA_BUCKET in envs
            and CondaRuntimeHandler.CONST_CONDA_OBJ_NAME in envs
        ):
            bucket = envs.pop(CondaRuntimeHandler.CONST_CONDA_BUCKET)
            namespace = envs.pop(CondaRuntimeHandler.CONST_CONDA_NAMESPACE)
            name = envs.pop(CondaRuntimeHandler.CONST_CONDA_OBJ_NAME)
            conda_spec = {
                CondaRuntime.CONST_CONDA_TYPE: envs.pop(
                    CondaRuntimeHandler.CONST_CONDA_TYPE
                ),
                CondaRuntime.CONST_CONDA_URI: f"oci://{bucket}@{namespace}/{name}",
            }
            if CondaRuntimeHandler.CONST_CONDA_REGION in envs:
                conda_spec[CondaRuntime.CONST_CONDA_REGION] = envs.pop(
                    CondaRuntimeHandler.CONST_CONDA_REGION
                )
            return conda_spec
        return None


class ScriptRuntimeHandler(CondaRuntimeHandler):
    """Runtime Handler for ScriptRuntime"""

    RUNTIME_CLASS = ScriptRuntime
    CONST_ENTRYPOINT = "JOB_RUN_ENTRYPOINT"

    def _translate_env(self, runtime: ScriptRuntime) -> dict:
        """Translate the environment variable.

        Parameters
        ----------
        runtime : ScriptRuntime
            An instance of ScriptRuntime

        Returns
        -------
        dict
            A dictionary contianing environment variables for OCI data science job.
        """
        envs = super()._translate_env(runtime)
        if runtime.entrypoint:
            envs[self.CONST_ENTRYPOINT] = runtime.entrypoint
        return envs

    def _translate_artifact(self, runtime: ScriptRuntime):
        return ScriptArtifact(runtime.source_uri, runtime)

    def _extract_envs(self, dsc_job) -> dict:
        """Extract the environment variables from data science job.
        ScriptRuntime may contain entrypoint as environment variable in addition to those for conda environment.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        spec = super()._extract_envs(dsc_job)
        envs = spec.pop(ScriptRuntime.CONST_ENV_VAR, {})
        entrypoint = envs.pop(ScriptRuntimeHandler.CONST_ENTRYPOINT, None)
        if entrypoint:
            spec[ScriptRuntime.CONST_ENTRYPOINT] = entrypoint
        if envs:
            spec[ScriptRuntime.CONST_ENV_VAR] = envs
        return spec

    def _extract_artifact(self, dsc_job):
        """Extract the job artifact from data science job.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        spec = super()._extract_artifact(dsc_job)
        spec.update({ScriptRuntime.CONST_SCRIPT_PATH: str(dsc_job.artifact)})
        return spec


class PythonRuntimeHandler(CondaRuntimeHandler):
    """Runtime Handler for PythonRuntime"""

    RUNTIME_CLASS = PythonRuntime
    PATH_DELIMITER = ":"
    CONST_JOB_ENTRYPOINT = "JOB_RUN_ENTRYPOINT"
    CONST_CODE_ENTRYPOINT = "CODE_ENTRYPOINT"
    CONST_ENTRY_FUNCTION = "ENTRY_FUNCTION"
    CONST_PYTHON_PATH = "PYTHON_PATH"
    CONST_OUTPUT_DIR = "OUTPUT_DIR"
    CONST_OUTPUT_URI = "OUTPUT_URI"
    CONST_WORKING_DIR = "WORKING_DIR"

    SPEC_MAPPINGS = {
        PythonRuntime.CONST_ENTRYPOINT: CONST_CODE_ENTRYPOINT,
        PythonRuntime.CONST_ENTRY_FUNCTION: CONST_ENTRY_FUNCTION,
        PythonRuntime.CONST_PYTHON_PATH: CONST_PYTHON_PATH,
        PythonRuntime.CONST_OUTPUT_DIR: CONST_OUTPUT_DIR,
        PythonRuntime.CONST_OUTPUT_URI: CONST_OUTPUT_URI,
        PythonRuntime.CONST_WORKING_DIR: CONST_WORKING_DIR,
    }

    def _translate_artifact(self, runtime: PythonRuntime):
        return PythonArtifact(runtime.script_uri, runtime)

    def _translate_env(self, runtime: PythonRuntime) -> dict:
        envs = super()._translate_env(runtime)
        envs.update(
            self._translate_specs(runtime, self.SPEC_MAPPINGS, self.PATH_DELIMITER)
        )

        if runtime.entrypoint:
            envs[self.CONST_CODE_ENTRYPOINT] = runtime.entrypoint
        else:
            envs[self.CONST_CODE_ENTRYPOINT] = os.path.basename(runtime.script_uri)

        envs[self.CONST_JOB_ENTRYPOINT] = PythonArtifact.CONST_DRIVER_SCRIPT
        return envs

    def _extract_envs(self, dsc_job) -> dict:
        """Extract the runtime specification from environment variables.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        spec = super()._extract_envs(dsc_job)
        envs = spec.pop(PythonRuntime.CONST_ENV_VAR, {})
        if self.CONST_CODE_ENTRYPOINT not in envs:
            raise IncompatibleRuntime()
        envs.pop(PythonRuntimeHandler.CONST_JOB_ENTRYPOINT)
        spec.update(self._extract_specs(envs, self.SPEC_MAPPINGS))
        if PythonRuntime.CONST_PYTHON_PATH in spec:
            spec[PythonRuntime.CONST_PYTHON_PATH] = spec[
                PythonRuntime.CONST_PYTHON_PATH
            ].split(self.PATH_DELIMITER)
        if envs:
            spec[PythonRuntime.CONST_ENV_VAR] = envs
        return spec


class NotebookRuntimeHandler(CondaRuntimeHandler):
    """Runtime Handler for NotebookRuntime"""

    RUNTIME_CLASS = NotebookRuntime
    CONST_NOTEBOOK_NAME = "JOB_RUN_NOTEBOOK"
    CONST_ENTRYPOINT = "JOB_RUN_ENTRYPOINT"
    CONST_OUTPUT_URI = "OUTPUT_URI"
    CONST_EXCLUDE_TAGS = "NOTEBOOK_EXCLUDE_TAGS"
    CONST_NOTEBOOK_ENCODING = "NOTEBOOK_ENCODING"

    SPEC_MAPPINGS = {
        NotebookRuntime.CONST_NOTEBOOK_PATH: CONST_NOTEBOOK_NAME,
        NotebookRuntime.CONST_OUTPUT_URI: CONST_OUTPUT_URI,
        NotebookRuntime.CONST_EXCLUDE_TAG: CONST_EXCLUDE_TAGS,
        NotebookRuntime.CONST_NOTEBOOK_ENCODING: CONST_NOTEBOOK_ENCODING,
    }

    def _translate_artifact(self, runtime: NotebookRuntime):
        return NotebookArtifact(runtime.notebook_uri, runtime)

    def _translate_env(self, runtime: NotebookRuntime) -> dict:
        envs = super()._translate_env(runtime)
        envs[self.CONST_NOTEBOOK_NAME] = os.path.basename(runtime.notebook_uri)
        envs[self.CONST_ENTRYPOINT] = NotebookArtifact.CONST_DRIVER_SCRIPT
        if runtime.notebook_encoding:
            envs[self.CONST_NOTEBOOK_ENCODING] = runtime.notebook_encoding
        if runtime.exclude_tag:
            envs[self.CONST_EXCLUDE_TAGS] = json.dumps(runtime.exclude_tag)
        if runtime.output_uri:
            envs[self.CONST_OUTPUT_URI] = runtime.output_uri
        return envs

    def _extract_envs(self, dsc_job) -> dict:
        """Extract the runtime specification from environment variables.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        spec = super()._extract_envs(dsc_job)
        envs = spec.pop(NotebookRuntime.CONST_ENV_VAR, {})
        if not (
            self.CONST_NOTEBOOK_NAME in envs
            and NotebookRuntimeHandler.CONST_ENTRYPOINT in envs
        ):
            raise IncompatibleRuntime()
        # Remove job run entrypoint since it is the same for notebook runtime.
        envs.pop(NotebookRuntimeHandler.CONST_ENTRYPOINT)
        # Extract exclude tags
        exclude_tags = envs.pop(NotebookRuntimeHandler.CONST_EXCLUDE_TAGS, None)
        if exclude_tags:
            # Exclude tags are in a JSON serialized string
            try:
                exclude_tags = json.loads(exclude_tags)
            except ValueError:
                # Ignore de-serialization error
                pass
            spec[NotebookRuntime.CONST_EXCLUDE_TAG] = exclude_tags

        spec.update(self._extract_specs(envs, self.SPEC_MAPPINGS))
        spec[NotebookRuntime.CONST_ENV_VAR] = envs
        return spec


class GitPythonRuntimeHandler(CondaRuntimeHandler):
    """Runtime Handler for GitPythonRuntime"""

    RUNTIME_CLASS = GitPythonRuntime

    PATH_DELIMITER = ":"
    CONST_GIT_URL = "GIT_URL"
    CONST_GIT_BRANCH = "GIT_BRANCH"
    CONST_GIT_COMMIT = "GIT_COMMIT"
    CONST_GIT_CODE_DIR = "CODE_DIR"
    CONST_GIT_SSH_SECRET_ID = "GIT_SECRET_OCID"
    CONST_SKIP_METADATA = "SKIP_METADATA_UPDATE"

    CONST_ENTRYPOINT = "GIT_ENTRYPOINT"
    CONST_ENTRY_FUNCTION = "ENTRY_FUNCTION"
    CONST_PYTHON_PATH = "PYTHON_PATH"
    CONST_OUTPUT_DIR = "OUTPUT_DIR"
    CONST_OUTPUT_URI = "OUTPUT_URI"
    CONST_WORKING_DIR = "WORKING_DIR"

    CONST_JOB_ENTRYPOINT = "JOB_RUN_ENTRYPOINT"

    SPEC_MAPPINGS = {
        GitPythonRuntime.CONST_GIT_URL: CONST_GIT_URL,
        GitPythonRuntime.CONST_BRANCH: CONST_GIT_BRANCH,
        GitPythonRuntime.CONST_COMMIT: CONST_GIT_COMMIT,
        GitPythonRuntime.CONST_ENTRYPOINT: CONST_ENTRYPOINT,
        GitPythonRuntime.CONST_ENTRY_FUNCTION: CONST_ENTRY_FUNCTION,
        GitPythonRuntime.CONST_PYTHON_PATH: CONST_PYTHON_PATH,
        GitPythonRuntime.CONST_GIT_SSH_SECRET_ID: CONST_GIT_SSH_SECRET_ID,
        GitPythonRuntime.CONST_OUTPUT_DIR: CONST_OUTPUT_DIR,
        GitPythonRuntime.CONST_OUTPUT_URI: CONST_OUTPUT_URI,
        GitPythonRuntime.CONST_WORKING_DIR: CONST_WORKING_DIR
    }

    def _translate_artifact(self, runtime: Runtime):
        """Specifies the driver script as the job artifact.
        runtime is not used in this method.

        Parameters
        ----------
        runtime : Runtime
            This is not used.

        Returns
        -------
        str
            Path to the git driver script.
        """
        return GitPythonArtifact()

    def _translate_env(self, runtime: GitPythonRuntime) -> dict:
        """Translate the environment variable.

        Parameters
        ----------
        runtime : GitPythonRuntime
            An instance of GitPythonRuntime

        Returns
        -------
        dict
            A dictionary containing environment variables for OCI data science job.
        """
        if not runtime.conda:
            raise ValueError(
                f"A conda pack is required for using the {runtime.__class__.__name__}. "
                "You can specify a service conda pack using with_service_conda()."
            )
        envs = super()._translate_env(runtime)
        envs.update(
            self._translate_specs(runtime, self.SPEC_MAPPINGS, self.PATH_DELIMITER)
        )
        if runtime.skip_metadata_update:
            envs[self.CONST_SKIP_METADATA] = "1"
        # Add entrypoint as the ADS driver is packed in a zip file.
        envs[self.CONST_JOB_ENTRYPOINT] = GitPythonArtifact.CONST_DRIVER_SCRIPT
        return envs

    def _extract_envs(self, dsc_job) -> dict:
        """Extract the environment variables from data science job.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        spec = super()._extract_envs(dsc_job)
        envs = spec.pop(CondaRuntime.CONST_ENV_VAR, {})

        if self.CONST_GIT_URL not in envs or self.CONST_ENTRYPOINT not in envs:
            raise IncompatibleRuntime()
        # Remove entrypoint as it's added by ADS
        envs.pop(self.CONST_JOB_ENTRYPOINT, None)
        spec.update(self._extract_specs(envs, self.SPEC_MAPPINGS))
        if GitPythonRuntime.CONST_PYTHON_PATH in spec:
            spec[GitPythonRuntime.CONST_PYTHON_PATH] = spec[
                GitPythonRuntime.CONST_PYTHON_PATH
            ].split(self.PATH_DELIMITER)
        if self.CONST_SKIP_METADATA in envs:
            envs.pop(self.CONST_SKIP_METADATA, None)
            spec[GitPythonRuntime.CONST_SKIP_METADATA] = True
        if envs:
            spec[ScriptRuntime.CONST_ENV_VAR] = envs
        return spec

    def _extract_artifact(self, dsc_job):
        """Git runtime uses the driver script as artifact. This will not be extracted."""
        return {}


class ContainerRuntimeHandler(RuntimeHandler):
    RUNTIME_CLASS = ContainerRuntime
    CMD_DELIMITER = ","
    CONST_CONTAINER_IMAGE = "CONTAINER_CUSTOM_IMAGE"
    CONST_CONTAINER_ENTRYPOINT = "CONTAINER_ENTRYPOINT"
    CONST_CONTAINER_CMD = "CONTAINER_CMD"

    def _translate_artifact(self, runtime: Runtime):
        """Specifies a dummy script as the job artifact.
        runtime is not used in this method.

        Parameters
        ----------
        runtime : Runtime
            This is not used.

        Returns
        -------
        str
            Path to the dummy script.
        """
        return os.path.join(
            os.path.dirname(__file__), "../../templates", "container.py"
        )

    def _translate_env(self, runtime: ContainerRuntime) -> dict:
        """Translate the environment variable.

        Parameters
        ----------
        runtime : GitPythonRuntime
            An instance of GitPythonRuntime

        Returns
        -------
        dict
            A dictionary containing environment variables for OCI data science job.
        """
        if not runtime.image:
            raise ValueError("Specify container image for ContainerRuntime.")
        envs = super()._translate_env(runtime)
        spec_mappings = {
            ContainerRuntime.CONST_IMAGE: self.CONST_CONTAINER_IMAGE,
            ContainerRuntime.CONST_ENTRYPOINT: self.CONST_CONTAINER_ENTRYPOINT,
            ContainerRuntime.CONST_CMD: self.CONST_CONTAINER_CMD,
        }
        envs.update(self._translate_specs(runtime, spec_mappings, self.CMD_DELIMITER))
        return envs

    @staticmethod
    def split_args(args: str) -> list:
        """Splits the cmd or entrypoint arguments for BYOC job into a list.
        BYOC jobs uses environment variables to store the values of cmd and entrypoint.
        In the values, comma(,) is used to separate cmd or entrypoint arguments.
        In YAML, the arguments are formatted into a list (Exec form).

        >>> ContainerRuntimeHandler.split_args("/bin/bash")
        ["/bin/bash"]
        >>> ContainerRuntimeHandler.split_args("-c,echo Hello World")
        ['-c', 'echo Hello World']

        Parameters
        ----------
        args : str
            Arguments in a comma separated string.

        Returns
        -------
        list
            Arguments in a list
        """
        if not args:
            return []
        return [
            arg.strip() for arg in args.split(ContainerRuntimeHandler.CMD_DELIMITER)
        ]

    def _extract_envs(self, dsc_job):
        """Extract the environment variables from data science job.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        dict
            A runtime specification dictionary for initializing a runtime.
        """
        spec = super()._extract_envs(dsc_job)
        envs = spec.pop(ContainerRuntime.CONST_ENV_VAR, {})
        if self.CONST_CONTAINER_IMAGE not in envs:
            raise IncompatibleRuntime()
        spec[ContainerRuntime.CONST_IMAGE] = envs.pop(self.CONST_CONTAINER_IMAGE)
        cmd = self.split_args(envs.pop(self.CONST_CONTAINER_CMD, ""))
        if cmd:
            spec[ContainerRuntime.CONST_CMD] = cmd
        entrypoint = self.split_args(envs.pop(self.CONST_CONTAINER_ENTRYPOINT, ""))
        if entrypoint:
            spec[ContainerRuntime.CONST_ENTRYPOINT] = entrypoint
        if envs:
            spec[ContainerRuntime.CONST_ENV_VAR] = envs
        return spec


class DataScienceJobRuntimeManager(RuntimeHandler):
    """This class is used by the DataScienceJob infrastructure to handle the runtime conversion.
    The translate() method determines the actual runtime handler by matching the RUNTIME_CLASS.
    The extract() method determines the actual runtime handler by checking if the runtime can be extracted.
    The order in runtime_handlers is used for extraction until a runtime is extracted.
    RuntimeHandler on the top of the list will have higher priority.
    If a runtime is a specify case of another runtime, the handler should be placed with higher priority.
    """

    runtime_handlers = [
        ContainerRuntimeHandler,
        GitPythonRuntimeHandler,
        NotebookRuntimeHandler,
        PythonRuntimeHandler,
        ScriptRuntimeHandler,
    ]

    def translate(self, runtime) -> dict:
        """Translates the runtime into a JSON payload for OCI API.
        This method determines the actual runtime handler by matching the RUNTIME_CLASS.

        Parameters
        ----------
        runtime : Runtime
            An instance of the runtime to be converted to a JSON payload.

        Returns
        -------
        dict
            JSON payload for defining a Data Science Job with OCI API
        """
        for runtime_handler in self.runtime_handlers:
            if runtime_handler.RUNTIME_CLASS == runtime.__class__:
                return runtime_handler(self.data_science_job).translate(runtime)
        raise NotImplementedError(
            f"{runtime.__class__.__name__} is not supported as the runtime of DataScienceJob."
        )

    def extract(self, dsc_job):
        """Extract the runtime from an OCI data science job object.

        This method determines the actual runtime handler by checking if the runtime can be extracted.

        Parameters
        ----------
        dsc_job : DSCJob or oci.datascience.models.Job
            The data science job containing runtime information.

        Returns
        -------
        Runtime:
            The runtime extracted from the data science job.
        """
        for runtime_handler in self.runtime_handlers:
            try:
                return runtime_handler(self.data_science_job).extract(dsc_job)
            except IncompatibleRuntime:
                pass
        raise NotImplementedError("Unable to extract runtime.")



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import copy

ALIAS_MAP = {
    "file_uri": "script_uri",
}


def normalize_config(config: dict) -> dict:
    """
    Normalize property names in a configuration so that they work directly with ADS.

    Parameters
    ----------
    config: dict
        input configuration, usually coming directly from an OCI response

    Returns
    -------
    dict
        output configuration
    """
    normalized = {}
    for k, v in config.items():
        k = ALIAS_MAP[k] if k in ALIAS_MAP else k
        if isinstance(v, dict):
            normalized[k] = normalize_config(v)
        else:
            normalized[k] = v
    return normalized


def get_value(obj, attr, default=None):
    """Gets a copy of the value from a nested dictionary of an object with nested attributes.

    Parameters
    ----------
    obj :
        An object or a dictionary
    attr :
        Attributes as a string seprated by dot(.)
    default :
        Default value to be returned if attribute is not found.

    Returns
    -------
    Any:
        A copy of the attribute value. For dict or list, a deepcopy will be returned.

    """
    keys = attr.split(".")
    val = default
    for key in keys:
        if hasattr(obj, key):
            val = getattr(obj, key)
        elif hasattr(obj, "get"):
            val = obj.get(key, default)
        else:
            return default
        obj = val
    return copy.deepcopy(val)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import copy
from typing import List

from ads.jobs import Job
from ads.jobs.builders.infrastructure.dsc_job import DataScienceJob
from ads.jobs.builders.runtimes.base import Runtime

PIPELINE_STEP_KIND_TO_OCI_MAP = {
    "dataScienceJob": "ML_JOB",
    "customScript": "CUSTOM_SCRIPT",
}

PIPELINE_STEP_KIND_FROM_OCI_MAP = {
    "ML_JOB": "dataScienceJob",
    "CUSTOM_SCRIPT": "customScript",
}

PIPELINE_STEP_KIND = {"dataScienceJob", "customScript"}

PIPELINE_STEP_RESTRICTED_CHAR_SET = {",", ">", "(", ")"}


class PipelineStep(Job):
    """Represents the Data Science Machine Learning Pipeline Step."""

    CONST_NAME = "name"
    CONST_JOB_ID = "jobId"
    CONST_DESCRIPTION = "description"
    CONST_DEPENDS_ON = "dependsOn"
    CONST_KIND = "stepType"
    CONST_MAXIMUM_RUNTIME_IN_MINUTES = "maximumRuntimeInMinutes"
    CONST_ENVIRONMENT_VARIABLES = "environmentVariables"
    CONST_COMMAND_LINE_ARGUMENTS = "commandLineArguments"
    CONST_STEP_INFRA_CONFIG_DETAILS = "stepInfrastructureConfigurationDetails"
    CONST_STEP_CONFIG_DETAILS = "stepConfigurationDetails"
    CONST_INFRASTRUCTURE = "infrastructure"
    CONST_RUNTIME = "runtime"

    def __init__(
        self,
        name: str,
        job_id: str = None,
        infrastructure=None,
        runtime=None,
        description=None,
        maximum_runtime_in_minutes=None,
        environment_variable=None,
        command_line_argument=None,
        kind=None,
    ) -> None:
        """Initialize a pipeline step.

        Parameters
        ----------
        name : str, required
            The name of the pipeline step.
        job_id : str, optional
            The job id of the pipeline step, by default None.
        infrastructure : Infrastructure, optional
            Pipeline step infrastructure, by default None.
        runtime : Runtime, optional
            Pipeline step runtime, by default None.
        description : str, optional
            The description for pipeline step, by default None.
        maximum_runtime_in_minutes : int, optional
            The maximum runtime in minutes for pipeline step, by default None.
        environment_variable : dict, optional
            The environment variable for pipeline step, by default None.
        command_line_argument : str, optional
            The command line argument for pipeline step, by default None.
        kind: str, optional
            The kind of pipeline step.

        Attributes
        ----------
        kind: str
            The kind of the object as showing in YAML.
        name: str
            The name of pipeline step.
        job_id: str
            The job id of pipeline step.
        infrastructure: DataScienceJob
            The infrastructure of pipeline step.
        runtime: Runtime
            The runtime of pipeline step.
        description: str
            The description of pipeline step.
        maximum_runtime_in_minutes: int
            The maximum runtime in minutes of pipeline step.
        environment_variable: dict
            The environment variables of pipeline step.
        argument: str
            The argument of pipeline step.
        depends_on: list
            The depends on of pipeline step.

        Methods
        -------
        with_job_id(self, job_id: str) -> PipelineStep
            Sets the job id for pipeline step.
        with_infrastructure(self, infrastructure) -> PipelineStep
            Sets the infrastructure for pipeline step.
        with_runtime(self, runtime) -> PipelineStep
            Sets the runtime for pipeline step.
        with_description(self, description: str) -> PipelineStep
            Sets the description for pipeline step.
        with_maximum_runtime_in_minutes(self, maximum_runtime_in_minutes: int) -> PipelineStep
            Sets the maximum runtime in minutes for pipeline step.
        with_environment_variable(self, **kwargs) -> PipelineStep
            Sets the environment variables for pipeline step.
        with_argument(self, *args, **kwargs) -> PipelineStep
            Sets the command line arguments for pipeline step.
        with_kind(self, kind: str) -> PipelineStep
            Sets the kind for pipeline step.
        to_dict(self) -> dict
            Serializes the pipeline step specification dictionary.
        from_dict(cls, config: dict) -> PipelineStep
            Initializes a PipelineStep from a dictionary containing the configurations.
        to_yaml(self, uri=None, **kwargs)
            Returns PipelineStep serialized as a YAML string
        from_yaml(cls, yaml_string=None, uri=None, **kwargs)
            Creates an PipelineStep from YAML string provided or from URI location containing YAML string

        Example
        -------
        Here is an example for defining a pipeline step using builder:

        .. code-block:: python

            from ads.pipeline import PipelineStep, CustomScriptStep, ScriptRuntime
            # Define an OCI Data Science pipeline step to run a python script
            pipeline_step = (
                PipelineStep(name="<pipeline_step_name>")
                .with_infrastructure(
                    CustomScriptStep()
                    .with_shape_name("VM.Standard2.1")
                    .with_block_storage_size(50)
                )
                .with_runtime(
                    ScriptRuntime()
                    .with_source("oci://bucket_name@namespace/path/to/script.py")
                    .with_service_conda("tensorflow26_p37_cpu_v2")
                    .with_environment_variable(ENV="value")
                    .with_argument("argument", key="value")
                    .with_maximum_runtime_in_minutes(200)
                )
            )

            # Another way to define an OCI Data Science pipeline step from existing job
            pipeline_step = (
                PipelineStep(name="<pipeline_step_name>")
                .with_job_id("<job_id>")
                .with_description("<description>")
            )

        See Also
        --------
        https://docs.oracle.com/en-us/iaas/tools/ads-sdk/latest/user_guide/pipeline/index.html
        """
        self.attribute_set = {
            "jobId",
            "stepType",
            "stepName",
            "description",
            "dependsOn",
            "stepInfrastructureConfigurationDetails",
            "stepConfigurationDetails",
        }

        super().__init__()
        if not name:
            raise ValueError("PipelineStep name must be specified.")
        elif any(char in PIPELINE_STEP_RESTRICTED_CHAR_SET for char in name):
            raise ValueError(
                "PipelineStep name can not include any of the "
                f"restricted characters in "
                f"{''.join(PIPELINE_STEP_RESTRICTED_CHAR_SET)}."
            )
        self.set_spec("name", name)

        if job_id:
            self.with_job_id(job_id)
        elif infrastructure and runtime:
            self.with_infrastructure(infrastructure)
            self.with_runtime(runtime)

        if maximum_runtime_in_minutes:
            self.with_maximum_runtime_in_minutes(maximum_runtime_in_minutes)
        if environment_variable:
            self.with_environment_variable(**environment_variable)
        if command_line_argument:
            self.with_argument(command_line_argument)
        if description:
            self.with_description(description)
        if kind:
            self.with_kind(kind)

    @property
    def name(self) -> str:
        """The name of pipeline step.

        Returns
        -------
        str
            The name of the pipeline step.
        """
        return self.get_spec(self.CONST_NAME)

    @property
    def kind(self) -> str:
        """The kind of the object as showing in YAML.

        Returns
        -------
        str
            The kind of the object as showing in YAML.
        """
        return self.get_spec(self.CONST_KIND)

    @property
    def job_id(self) -> str:
        """The job id of the pipeline step.

        Returns
        -------
        str
            The job id of the pipeline step.
        """
        return self.get_spec(self.CONST_JOB_ID)

    def with_job_id(self, job_id: str) -> "PipelineStep":
        """Sets the job id for pipeline step.

        Parameters
        ----------
        job_id : str
            The job id of pipeline step.

        Returns
        -------
            Pipeline step instance (self).
        """
        if not self.kind:
            self.set_spec(self.CONST_KIND, "ML_JOB")
        return self.set_spec(self.CONST_JOB_ID, job_id)

    @property
    def infrastructure(self) -> "DataScienceJob":
        """The infrastructure of the pipeline step.

        Returns
        -------
        DataScienceJob :
            Data science pipeline step instance.
        """
        return self.get_spec(self.CONST_INFRASTRUCTURE)

    def with_infrastructure(self, infrastructure) -> "PipelineStep":
        """Sets the infrastructure for pipeline step.

        Parameters
        ----------
        infrastructure :
            The infrastructure of pipeline step.

        Returns
        -------
            Pipeline step instance (self).
        """
        if not self.kind:
            self.set_spec(self.CONST_KIND, "CUSTOM_SCRIPT")
        return self.set_spec(self.CONST_INFRASTRUCTURE, infrastructure)

    @property
    def runtime(self) -> "Runtime":
        """The runtime of the pipeline step.

        Returns
        -------
        Runtime :
            Runtime instance.
        """
        return self.get_spec(self.CONST_RUNTIME)

    def with_runtime(self, runtime) -> "PipelineStep":
        """Sets the runtime for pipeline step.

        Parameters
        ----------
        runtime :
            The runtime of pipeline step.

        Returns
        -------
            Pipeline step instance (self).
        """
        if not self.kind:
            self.set_spec(self.CONST_KIND, "CUSTOM_SCRIPT")
        return self.set_spec(self.CONST_RUNTIME, runtime)

    @property
    def description(self) -> str:
        """The description of the pipeline step.

        Returns
        -------
        str
            The description of the pipeline step.
        """
        return self.get_spec(self.CONST_DESCRIPTION)

    def with_description(self, description: str) -> "PipelineStep":
        """Sets the description for pipeline step.

        Parameters
        ----------
        description : str
            The description of pipeline step.

        Returns
        -------
            Pipeline step instance (self).
        """
        return self.set_spec(self.CONST_DESCRIPTION, description)

    @property
    def maximum_runtime_in_minutes(self) -> int:
        """The maximum runtime in minutes of pipeline step.

        Returns
        -------
        int
            The maximum runtime in minutes of the pipeline step.
        """
        return self.get_spec(self.CONST_MAXIMUM_RUNTIME_IN_MINUTES)

    def with_maximum_runtime_in_minutes(
        self, maximum_runtime_in_minutes: int
    ) -> "PipelineStep":
        """Sets the maximum runtime in minutes of pipeline step.

        Parameters
        ----------
        maximum_runtime_in_minutes : int
            The maximum runtime in minutes of pipeline step.

        Returns
        -------
            Pipeline step instance (self).
        """
        return self.set_spec(
            self.CONST_MAXIMUM_RUNTIME_IN_MINUTES, maximum_runtime_in_minutes
        )

    @property
    def environment_variable(self) -> dict:
        """The environment variables of the pipeline step.

        Returns
        -------
        dict:
            The environment variables of the pipeline step.
        """
        return self.get_spec(self.CONST_ENVIRONMENT_VARIABLES)

    def with_environment_variable(self, **kwargs) -> "PipelineStep":
        """Sets environment variables of the pipeline step.

        Parameters
        ----------
        kwargs:
            Keyword arguments.
            To add a keyword argument without value, set the value to None.

        Returns
        -------
        Pipeline
            The Pipeline step instance (self).
        """
        if kwargs:
            environment_variable_dict = {}
            for k, v in kwargs.items():
                environment_variable_dict[k] = v
            self.set_spec(self.CONST_ENVIRONMENT_VARIABLES, environment_variable_dict)
        return self

    @property
    def argument(self) -> str:
        """The command line arguments of the pipeline step.

        Returns
        -------
        str:
            The command line arguments of the pipeline step.
        """
        return self.get_spec(self.CONST_COMMAND_LINE_ARGUMENTS)

    def with_argument(self, *args, **kwargs) -> "PipelineStep":
        """Adds command line arguments to the pipeline step.
        Existing arguments will be preserved.
        This method can be called (chained) multiple times to add various arguments.
        For example, pipeline.with_argument(key="val").with_argument("path/to/file") will result in:
        "--key val path/to/file"

        Parameters
        ----------
        args:
            Positional arguments.
            In a single method call, positional arguments are always added before keyword arguments.
            You can call with_argument() to add positional arguments after keyword arguments.

        kwargs:
            Keyword arguments.
            To add a keyword argument without value, set the value to None.

        Returns
        -------
        Pipeline
            The Pipeline step instance (self).

        Raises
        ------
        ValueError
            Keyword arguments with space in a key.
        """
        arg_values = self.get_spec(self.CONST_COMMAND_LINE_ARGUMENTS, [])
        args = [str(arg) for arg in args]
        arg_values.extend(args)
        for k, v in kwargs.items():
            if " " in k:
                raise ValueError("Argument key %s cannot contain space.", str(k))
            arg_values.append(f"--{str(k)}")
            # Ignore None value
            if v is None:
                continue
            arg_values.append(str(v))
        arg_string = " ".join(arg_values)
        self.set_spec(self.CONST_COMMAND_LINE_ARGUMENTS, arg_string)
        return self

    @property
    def depends_on(self) -> list:
        """The list of upstream pipeline steps for (self).

        Returns
        -------
        list
            The list of upstream pipeline steps for (self).
        """
        return self.get_spec(self.CONST_DEPENDS_ON)

    def _with_depends_on(self, depends_on: List["PipelineStep"]) -> "PipelineStep":
        """Sets the list of upstream pipeline steps for (self).

        Parameters
        ----------
        depends_on : list of PipelineStep objects
            The list of pipeline steps that (self) depends on.

        Returns
        -------
            Pipeline step instance (self).
        """
        if not depends_on:
            return self.set_spec(self.CONST_DEPENDS_ON, [])

        step_list = []
        for step in depends_on:
            step_list.append(step.name)
        return self.set_spec(self.CONST_DEPENDS_ON, step_list)

    def with_kind(self, kind: str) -> "PipelineStep":
        """Sets the kind of pipeline step.

        Parameters
        ----------
        kind : str
            The kind of pipeline step.

        Returns
        -------
            Pipeline step instance (self).
        """
        if kind in PIPELINE_STEP_KIND:
            self.set_spec(self.CONST_KIND, PIPELINE_STEP_KIND_TO_OCI_MAP[kind])
        else:
            raise ValueError(
                "Invalid PipelineStep kind. The allowed "
                f"values are {', '.join(PIPELINE_STEP_KIND)}."
            )
        return self.set_spec(self.CONST_KIND, kind)

    def to_dict(self) -> dict:
        """Serializes the pipeline step specification dictionary.

        Returns
        -------
        dict
            A dictionary containing pipeline step specification.
        """
        dict_details = copy.deepcopy(super().to_dict())
        if self.kind in PIPELINE_STEP_KIND_FROM_OCI_MAP:
            dict_details["kind"] = PIPELINE_STEP_KIND_FROM_OCI_MAP[self.kind]

        # remove information not going to show in to_dict()
        if self.CONST_INFRASTRUCTURE in dict_details["spec"]:
            dict_details["spec"][self.CONST_INFRASTRUCTURE].pop("type", None)
            dict_details["spec"][self.CONST_INFRASTRUCTURE]["spec"].pop(
                "jobInfrastructureType", None
            )
            dict_details["spec"][self.CONST_INFRASTRUCTURE]["spec"].pop("jobType", None)

        if self.job_id:
            dict_details["spec"][self.CONST_JOB_ID] = self.job_id
        if self.description:
            dict_details["spec"][self.CONST_DESCRIPTION] = self.description

        dict_details["spec"].pop(self.CONST_DEPENDS_ON, None)

        return dict_details

    @classmethod
    def from_dict(cls, config: dict) -> "PipelineStep":
        """Initializes a PipelineStep from a dictionary containing the configurations.

        Parameters
        ----------
        config : dict
            A dictionary containing the infrastructure and runtime specifications.

        Returns
        -------
        PipelineStep
            A PipelineStep instance

        Raises
        ------
        NotImplementedError
            If the type of the intrastructure or runtime is not supported.
        """
        if not isinstance(config, dict):
            raise ValueError("The config data for initializing the job is invalid.")
        spec = config.get("spec")

        mappings = {
            "infrastructure": cls._INFRASTRUCTURE_MAPPING,
            "runtime": cls._RUNTIME_MAPPING,
        }
        if spec["name"]:
            pipeline_step = cls(name=spec["name"])
        else:
            raise ValueError("PipelineStep name must be specified.")

        if config.get("kind", None):
            step_kind = config.get("kind")

            if step_kind in PIPELINE_STEP_KIND:
                pipeline_step.set_spec(
                    cls.CONST_KIND, PIPELINE_STEP_KIND_TO_OCI_MAP[step_kind]
                )
            else:
                raise ValueError(
                    "Invalid PipelineStep kind. The allowed "
                    f"values are {', '.join(PIPELINE_STEP_KIND)}."
                )
        else:
            pipeline_step.set_spec(cls.CONST_KIND, "ML_JOB")

        for key, value in spec.items():
            if key in mappings:
                mapping = mappings[key]
                child_config = copy.deepcopy(value)
                if key == "infrastructure":
                    child_config["type"] = "dataScienceJob"
                if child_config.get("type") not in mapping:
                    raise NotImplementedError(
                        f"{key.title()} type: {child_config.get('type')} is not supported."
                    )
                pipeline_step.set_spec(
                    key, mapping[child_config.get("type")].from_dict(child_config)
                )
            elif key == cls.CONST_STEP_CONFIG_DETAILS:
                for attribute in value:
                    pipeline_step.set_spec(attribute, value[attribute])
            else:
                pipeline_step.set_spec(key, value)

        return pipeline_step



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from dataclasses import dataclass, field
from typing import Dict

from ads.common.serializer import DataClassSerializable
from ads.model.runtime.env_info import InferenceEnvInfo


@dataclass(repr=False)
class ModelDeploymentDetails(DataClassSerializable):
    """ModelDeploymentDetails class."""

    inference_conda_env: InferenceEnvInfo = field(default_factory=InferenceEnvInfo)

    @staticmethod
    def _validate_dict(obj_dict: Dict) -> bool:
        """Validate the content in the ditionary format from the yaml file.

        Parameters
        ----------
        obj_dict: (Dict)
            yaml file content to validate.

        Returns
        -------
        bool
            Validation result.
        """
        assert obj_dict and (
            "INFERENCE_CONDA_ENV" in obj_dict
        ), "`model_deployment_details` must have `INFERENCE_CONDA_ENV` field."
        return True



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import warnings

warnings.warn(
    (
        "The `ads.catalog.notebook` is deprecated in `oracle-ads 2.6.9` and will be removed in `oracle-ads 3.0`."
    ),
    DeprecationWarning,
    stacklevel=2,
)

from pandas import DataFrame
import oci
from oci.data_science.models import (
    NotebookSessionSummary,
    UpdateNotebookSessionDetails,
    CreateNotebookSessionDetails,
    NotebookSession,
    NotebookSessionConfigurationDetails,
)
from oci.exceptions import ServiceError
from types import MethodType

from ads.catalog.summary import SummaryList
from ads.common import utils
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common import auth as authutil
from ads.common import oci_client as oc
from ads.config import (
    OCI_IDENTITY_SERVICE_ENDPOINT,
    NB_SESSION_COMPARTMENT_OCID,
    PROJECT_OCID,
    OCI_ODSC_SERVICE_ENDPOINT,
)

create_notebook_details_attributes = CreateNotebookSessionDetails().swagger_types.keys()
notebook_attributes = list(NotebookSession().swagger_types.keys())
notebook_attributes.append("user_name")
update_notebook_details_attributes = UpdateNotebookSessionDetails().swagger_types.keys()


class NotebookSummaryList(SummaryList):
    def __init__(self, notebook_list, response=None, datetime_format=utils.date_format):
        super(NotebookSummaryList, self).__init__(
            notebook_list, datetime_format=datetime_format
        )
        self.response = response
        self.df["notebook_session_url"] = self.df[
            self.df["lifecycle_state"] == "ACTIVE"
        ]["notebook_session_url"].apply(
            lambda x: "<a href='%s'>%s</a>"
            % (x if x.startswith("http") else "http://%s" % x, "open")
        )

    def __add__(self, rhs):
        return NotebookSummaryList(
            list.__add__(self, rhs), datetime_format=self.datetime_format
        )

    def sort_by(self, columns, reverse=False):
        """
        Performs a multi-key sort on a particular set of columns and returns the sorted NotebookSummaryList
        Results are listed in a descending order by default.

        Parameters
        ----------
        columns: List of string
          A list of columns which are provided to sort on
        reverse: Boolean (defaults to false)
          If you'd like to reverse the results (for example, to get ascending instead of descending results)

        Returns
        -------
        NotebookSummaryList: A sorted NotebookSummaryList
        """
        return NotebookSummaryList(
            self._sort_by(columns, reverse=reverse),
            datetime_format=self.datetime_format,
        )

    def filter(self, selection, instance=None):
        """
        Filter the notebook list according to a lambda filter function, or list comprehension.

        Parameters
        ----------
        selection: lambda function filtering notebook instances, or a list-comprehension
            function of list filtering notebooks
        instance: list, optional
            list to filter, optional, defaults to self

        Raises
        ------
        ValueError: If selection passed is not correct. For example: selection=oci.data_science.models.NotebookSession.
        """
        instance = instance if instance is not None else self

        if callable(selection):
            res = list(filter(selection, instance))
            # lambda filtering
            if len(res) == 0:
                print("No notebook sessions found")
                return
            return NotebookSummaryList(res)
        elif isinstance(selection, list):
            # list comprehension
            if len(selection) == 0:
                print("No notebook sessions found")
                return
            return NotebookSummaryList(selection, datetime_format=self.datetime_format)
        else:
            raise ValueError(
                "Filter selection must be a function or a NotebookSummaryList"
            )


class NotebookCatalog:
    def __init__(self, compartment_id=None):
        self.compartment_id = (
            NB_SESSION_COMPARTMENT_OCID if compartment_id is None else compartment_id
        )
        if self.compartment_id is None:
            raise ValueError("compartment_id needs to be specified.")

        if OCI_ODSC_SERVICE_ENDPOINT:
            odsc_auth = authutil.default_signer(
                client_kwargs={"service_endpoint": OCI_ODSC_SERVICE_ENDPOINT}
            )
        else:
            odsc_auth = authutil.default_signer()
        self.ds_client = oc.OCIClientFactory(**odsc_auth).data_science

        if OCI_IDENTITY_SERVICE_ENDPOINT:
            identity_auth = authutil.default_signer(
                client_kwargs={"service_endpoint": OCI_IDENTITY_SERVICE_ENDPOINT}
            )
        else:
            identity_auth = authutil.default_signer()
        self.identity_client = oc.OCIClientFactory(**identity_auth).identity

        self.short_id_index = {}

    def __getitem__(self, key):
        return self.get_notebook_session(key)

    def __iter__(self):
        return self.list_notebook_session().__iter__()

    def __len__(self):
        return len(self.list_notebook_session())

    def _decorate_notebook_session(self, notebook, response=None):
        notebook.catalog = self
        notebook.response = response
        notebook.user_name = ""
        notebook.swagger_types["user_name"] = "str"
        notebook.swagger_types["user_email"] = "str"
        try:
            user = self.identity_client.get_user(notebook.created_by)
            notebook.user = user.data
            notebook.user_name = user.data.name
            notebook.user_email = user.data.email
        except:
            pass

        def commit(notebook_self, **kwargs):
            # exclude 'notebook_session_configuration_details' key because it's can't be committed/changed
            keys_to_commit = ["display_name", "freeform_tags", "defined_tags"]
            update_notebook_details = UpdateNotebookSessionDetails(
                **{key: getattr(notebook, key) for key in keys_to_commit}
            )
            return self.update_notebook_session(
                notebook_self.id, update_notebook_details, **kwargs
            )

        def rollback(notebook_self):
            """
            Rollback the project to a remote state

            Returns
            -------
            None
            """
            notebook_self.__dict__.update(
                self.get_notebook_session(notebook_self.id).__dict__
            )

        def to_dataframe(notebook_self):
            df = DataFrame.from_dict(
                {
                    key: getattr(notebook_self, key)
                    for key in notebook_attributes
                    if hasattr(notebook_self, key)
                },
                orient="index",
                columns=[""],
            )
            return df

        @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
        def show_in_notebook(notebook_self):
            """
            Describe the project by showing it's properties
            """
            from IPython.core.display import display

            display(notebook_self)

        def _repr_html_(notebook_self):
            df = notebook_self.to_dataframe()
            df.loc["notebook_session_url"] = df.loc["notebook_session_url"].apply(
                lambda x: "<a href='%s'>%s</a>"
                % (x if x.startswith("http") else "http://%s" % x, "open")
            )
            return df.style.set_properties(**{"margin-left": "0px"}).render()

        notebook.commit = MethodType(commit, notebook)
        notebook.rollback = MethodType(rollback, notebook)
        notebook.to_dataframe = MethodType(to_dataframe, notebook)
        notebook.show_in_notebook = MethodType(show_in_notebook, notebook)
        notebook._repr_html_ = MethodType(_repr_html_, notebook)

        return notebook

    def list_notebook_session(
        self, include_deleted=False, datetime_format=utils.date_format, **kwargs
    ):
        """
        List all notebooks in a given compartment

        Parameters
        ----------
        include_deleted: bool, optional, default=False
            Whether to include deleted notebooks in the returned list
        datetime_format: str, optional, default: '%Y-%m-%d %H:%M:%S'
            Change format for date time fields

        Returns
        -------
        NotebookSummaryList: A List of notebooks.

        Raises
        ------
        KeyError: If the resource was not found or do not have authorization to access that resource.
        """
        try:
            list_notebook_response = self.ds_client.list_notebook_sessions(
                self.compartment_id, **kwargs
            )
            if (
                list_notebook_response.data is None
                or len(list_notebook_response.data) == 0
            ):
                print("No notebooks found.")
                return
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise

        notebook_list_filtered = [
            self._decorate_notebook_session(notebook)
            for notebook in list_notebook_response.data
            if include_deleted
            or notebook.lifecycle_state
            != NotebookSessionSummary.LIFECYCLE_STATE_DELETED
        ]

        # handle empty list
        if notebook_list_filtered is None or len(notebook_list_filtered) == 0:
            print("No notebooks found.")
            return

        nsl = NotebookSummaryList(
            notebook_list_filtered,
            list_notebook_response,
            datetime_format=datetime_format,
        )
        self.short_id_index.update(nsl.short_id_index)
        return nsl

    def create_notebook_session(
        self,
        display_name=None,
        project_id=None,
        shape=None,
        block_storage_size_in_gbs=None,
        subnet_id=None,
        **kwargs
    ):
        """
        Create a new notebook session with the supplied details.

        Parameters
        ----------
        display_name: str, required
            The value to assign to the display_name property of this CreateNotebookSessionDetails.
        project_id: str, required
            The value to assign to the project_id property of this CreateNotebookSessionDetails.
        shape: str, required
            The value to assign to the shape property of this NotebookSessionConfigurationDetails.
            Allowed values for this property are: "VM.Standard.E2.2", "VM.Standard.E2.4",
            "VM.Standard.E2.8", "VM.Standard2.1", "VM.Standard2.2", "VM.Standard2.4", "VM.Standard2.8",
            "VM.Standard2.16","VM.Standard2.24".
        block_storage_size_in_gbs: int, required
            Size of the block storage drive. Limited to values between 50 (GB) and 1024 (1024GB = 1TB)
        subnet_id: str, required
            The OCID of the subnet resource where the notebook is to be created.
        kwargs: dict, optional
            Additional kwargs passed to `DataScienceClient.create_notebook_session()`

        Returns
        -------
        oci.data_science.models.NotebookSession: A new notebook record.

        Raises
        ------
        KeyError: If the resource was not found or do not have authorization to access that resource.
        """
        notebook_session_configuration_details = NotebookSessionConfigurationDetails(
            shape=shape,
            block_storage_size_in_gbs=block_storage_size_in_gbs,
            subnet_id=subnet_id,
        )
        project_id = PROJECT_OCID if project_id is None else project_id
        create_notebook_details = CreateNotebookSessionDetails(
            display_name=display_name,
            project_id=project_id,
            compartment_id=self.compartment_id,
            notebook_session_configuration_details=notebook_session_configuration_details,
        )
        try:
            create_notebook_response = self.ds_client.create_notebook_session(
                create_notebook_details, **kwargs
            )
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise

        return self.get_notebook_session(create_notebook_response.data.id)

    def get_notebook_session(self, notebook_id):
        """
        Get the notebook based on notebook_id

        Parameters
        ----------
        notebook_id: str, required
            The OCID of the notebook to get.

        Returns
        -------
        oci.data_science.models.NotebookSession: The oci.data_science.models.NotebookSession with the matching ID.

        Raises
        ------
        KeyError: If the resource was not found or do not have authorization to access that resource.
        """
        if not notebook_id.startswith("ocid"):
            notebook_id = self.short_id_index[notebook_id]

        try:
            get_notebook_response = self.ds_client.get_notebook_session(notebook_id)
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise
        return self._decorate_notebook_session(
            get_notebook_response.data, get_notebook_response
        )

    def update_notebook_session(
        self, notebook_id, update_notebook_details=None, **kwargs
    ):
        """
        Updates a notebook with given notebook_id, using the provided update data

        Parameters
        ----------
        notebook_id: str
            notebook_id OCID to update
        update_notebook_details: oci.data_science.models.UpdateNotebookSessionDetails
            contains the new notebook details data to apply
        kwargs: dict, optional
            Update notebook session details can be supplied instead as kwargs

        Returns
        -------
        oci.data_science.models.NotebookSession: The updated Notebook record

        Raises
        ------
        KeyError: If the resource was not found or do not have authorization to access that resource.
        """
        if not notebook_id.startswith("ocid"):
            notebook_id = self.short_id_index[notebook_id]
        if update_notebook_details is None:
            update_notebook_details = UpdateNotebookSessionDetails(
                **{
                    k: v
                    for k, v in kwargs.items()
                    if k in update_notebook_details_attributes
                }
            )
            update_notebook_details.compartment_id = self.compartment_id
            # filter kwargs removing used keys
            kwargs = {
                k: v
                for k, v in kwargs.items()
                if k not in update_notebook_details_attributes
            }
        try:
            update_notebook_response = self.ds_client.update_notebook_session(
                notebook_id, update_notebook_details, **kwargs
            )
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise
        return self._decorate_notebook_session(
            update_notebook_response.data, update_notebook_response
        )

    def delete_notebook_session(self, notebook, **kwargs):
        """
        Deletes the notebook based on notebook_id.

        Parameters
        ----------
        notebook: str ID or oci.data_science.models.NotebookSession,required
            The OCID of the notebook to delete as a string, or a Notebook Session instance

        Returns
        -------
        Bool: True if delete was successful, false otherwise
        """
        try:
            notebook_id = (
                notebook.id
                if isinstance(notebook, NotebookSession)
                else self.short_id_index[notebook]
                if not notebook.startswith("ocid")
                else notebook
            )
            self.ds_client.delete_notebook_session(notebook_id, **kwargs)
            return True
        except:
            return False



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
import time
import sys
import warnings
from abc import ABC, abstractmethod, abstractproperty
import math
import pandas as pd
import numpy as np
from sklearn import set_config
from sklearn.dummy import DummyClassifier, DummyRegressor

import matplotlib.pyplot as plt

import ads
from ads.common.utils import (
    ml_task_types,
    wrap_lines,
    is_documentation_mode,
    is_notebook,
)
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common.decorator.deprecate import deprecated
from ads.dataset.label_encoder import DataFrameLabelEncoder
from ads.dataset.helper import is_text_data

from ads.common import logger, utils


class AutoMLProvider(ABC):
    """
    Abstract Base Class defining the structure of an AutoML solution. The solution needs to
    implement train() and get_transformer_pipeline().
    """

    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self):
        self.X_train = None
        self.y_train = None
        self.X_valid = None
        self.y_valid = None
        self.client = None
        self.ml_task_type = None
        self.class_names = None
        self.transformer_pipeline = None
        self.est = None

    def setup(
        self,
        X_train,
        y_train,
        ml_task_type,
        X_valid=None,
        y_valid=None,
        class_names=None,
        client=None,
    ):
        """
        Setup arguments to the AutoML instance.

        Parameters
        ----------
        X_train : DataFrame
            Training features
        y_train : DataFrame
            Training labels
        ml_task_type : One of ml_task_type.{REGRESSION,BINARY_CLASSIFICATION,
            MULTI_CLASS_CLASSIFICATION,BINARY_TEXT_CLASSIFICATION,MULTI_CLASS_TEXT_CLASSIFICATION}
        X_valid : DataFrame
            Validation features
        y_valid : DataFrame
            Validation labels
        class_names : list
            Unique values in y_train
        client : object
            Dask client instance for distributed execution
        """
        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid
        self.ml_task_type = ml_task_type
        self.client = client
        self.class_names = class_names

    @property
    def est(self):
        """
        Returns the estimator.

        The estimator can be a standard sklearn estimator or any object that implement methods from
        (BaseEstimator, RegressorMixin) for regression or (BaseEstimator, ClassifierMixin) for classification.

        Returns
        -------
        est : An instance of estimator
        """
        return self.__est

    @est.setter
    def est(self, est):
        self.__est = est

    @abstractmethod
    def train(self, **kwargs):
        """
        Calls fit on estimator.

        This method is expected to set the 'est' property.

        Parameters
        ----------
        kwargs: dict, optional
        kwargs to decide the estimator and arguments for the fit method
        """
        pass

    @abstractmethod
    def get_transformer_pipeline(self):
        """
        Returns a list of transformers representing the transformations done on data before model prediction.

        This method is optional to implement, and is used only for visualizing transformations on data using
        ADSModel#visualize_transforms().

        Returns
        -------
        transformers_list : list of transformers implementing fit and transform
        """
        pass


class BaselineModel(object):
    """
    A BaselineModel object that supports fit/predict/predict_proba/transform
    interface. Labels (y) are encoded using DataFrameLabelEncoder.
    """

    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self, est):
        self.est = est
        self.df_label_encoder = DataFrameLabelEncoder()

    def predict(self, X):

        """
        Runs the Baselines predict function and returns the result.

        Parameters
        ----------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be predicted on

        Returns
        -------
        List: A list of predictions performed on the input data.
        """

        X = self.transform(X)
        return self.est.predict(X)

    def predict_proba(self, X):

        """
        Runs the Baselines predict_proba function and returns the result.

        Parameters
        ----------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be predicted on

        Returns
        -------
        List: A list of probabilities of being part of a class
        """

        X = self.transform(X)
        return self.est.predict_proba(X)

    def fit(self, X, y):

        """
        Fits the baseline estimator.

        Parameters
        ----------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be predicted on
        Y: Dataframe, Series, or list-like
          A Dataframe, series, or list-like object holding the labels


        Returns
        -------
        estimator: The fitted estimator
        """

        self.est.fit(X, y)
        return self

    def transform(self, X):

        """
        Runs the Baselines transform function and returns the result.

        Parameters
        ---------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be transformed

        Returns
        -------
        Dataframe or list-like: The transformed Dataframe. Currently, no transformation is performed by the default Baseline Estimator.
        """

        return X

    def __getattr__(self, item):
        return getattr(self.est, item)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state

    def __repr__(self):
        set_config()
        return str(self.est)[:-2]


class BaselineAutoMLProvider(AutoMLProvider):
    def get_transformer_pipeline(self):
        """
        Returns a list of transformers representing the transformations done on data before model prediction.

        This method is used only for visualizing transformations on data using
        ADSModel#visualize_transforms().

        Returns
        -------
        transformers_list : list of transformers implementing fit and transform
        """
        msg = "Baseline"
        return [("automl_preprocessing", AutoMLPreprocessingTransformer(msg))]

    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self, est):
        """
        Generates a baseline model using the Zero Rule algorithm by default. For a classification
        predictive modeling problem where a categorical value is predicted, the Zero
        Rule algorithm predicts the class value that has the most observations in the training dataset.

        Parameters
        ----------
        est : BaselineModel
            An estimator that supports the fit/predict/predict_proba interface.
            By default, DummyClassifier/DummyRegressor are used as estimators
        """
        super(BaselineAutoMLProvider, self).__init__()
        self.est = est

    def __repr__(self):
        set_config()
        return str(self.est)[:-2]

    def train(self, **kwargs):
        self.est = self.decide_estimator(**kwargs)
        if self.est is None:
            raise ValueError(
                "Baseline model for (%s) is not supported" % self.ml_task_type
            )
        try:
            self.est.fit(self.X_train, self.y_train)
        except Exception as e:
            warning_message = f"The baseline estimator failed to fit the data. It could not evaluate {self.est} and gave the exception {e}."
            logger.warning(warning_message)

    def decide_estimator(self, **kwargs):
        """
        Decides which type of BaselineModel to generate.

        Returns
        -------
        Modell: BaselineModel
            A baseline model generated for the particular ML task being performed
        """
        if self.est is not None:
            return self.est
        else:
            if self.ml_task_type == ml_task_types.REGRESSION:
                return BaselineModel(DummyRegressor())
            elif self.ml_task_type in [
                ml_task_types.BINARY_CLASSIFICATION,
                ml_task_types.MULTI_CLASS_CLASSIFICATION,
                ml_task_types.BINARY_TEXT_CLASSIFICATION,
                ml_task_types.MULTI_CLASS_TEXT_CLASSIFICATION,
            ]:
                return BaselineModel(DummyClassifier())


# An installation of oracle labs automl is required only for this class
class OracleAutoMLProvider(AutoMLProvider, ABC):
    @deprecated(
        "2.6.7",
        details="Oracle AutoML is recommended to be directly instantiated by importing automlx package",
        raise_error=True,
    )
    def __init__(
        self, n_jobs=-1, loglevel=None, logger_override=None, model_n_jobs: int = 1
    ):
        """
        The Oracle AutoML Provider automatically provides a tuned ML pipeline that best models the given a training
        dataset and a prediction task at hand.

        Parameters
        ----------
        n_jobs : int
            Specifies the degree of parallelism for Oracle AutoML. -1 (default) means that AutoML will use all
            available cores.
        loglevel : int
            The verbosity of output for Oracle AutoML. Can be specified using the Python logging module
            (https://docs.python.org/3/library/logging.html#logging-levels).
        model_n_jobs: (optional, int). Defaults to 1.
            Specifies the model parallelism used by AutoML.
            This will be passed to the underlying model it is training.
        """
        try:
            self.automl = __import__("automl")
            self.cpuinfo = __import__("cpuinfo")
        except ModuleNotFoundError as e:
            utils._log_missing_module("automl", "ads[labs]")
            raise e
        super(OracleAutoMLProvider, self).__init__()
        if loglevel is None:
            loglevel = logging.DEBUG if ads.debug_mode else logging.ERROR

        self.automl.init(
            engine="local",
            engine_opts={"n_jobs": n_jobs, "model_n_jobs": model_n_jobs},
            logger=logger_override,
            loglevel=loglevel,
        )

    def __repr__(self):
        super(OracleAutoMLProvider, self).__repr__()

    def get_transformer_pipeline(self):
        """
        Returns a list of transformers representing the transformations done on data before model prediction.

        This method is used only for visualizing transformations on data using
        ADSModel#visualize_transforms().

        Returns
        -------
        transformers_list : list of transformers implementing fit and transform
        """
        if hasattr(self.est, "text") and not self.est.text:
            msg1 = wrap_lines(
                self.est.selected_features_names_, heading="Select features:"
            )
            return [("automl_feature_selection", AutoMLFeatureSelection(msg1))]
        else:
            msg = "Apply Tfidf Vectorization\n"
            msg += "Normalize features\n"
            msg += "Label encode target"
            return [("automl_preprocessing", AutoMLPreprocessingTransformer(msg))]

    def selected_model_name(self):
        """
        Return the name of the selected model by AutoML.
        """
        return self.est.selected_model_

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def print_summary(
        self,
        max_rows=None,
        sort_column="Mean Validation Score",
        ranking_table_only=False,
    ):
        """
        Prints a summary of the Oracle AutoML Pipeline in the last train() call.

        Parameters
        ----------
        max_rows : int
            Number of trials to print. Pass in None to print all trials
        sort_column: string
            Column to sort results by. Must be one of ['Algorithm', '#Samples', '#Features', 'Mean Validation Score',
            'Hyperparameters', 'All Validation Scores', 'CPU Time']
        ranking_table_only: bool
            Table to be displayed. Pass in False to display the complete table.
            Pass in True to display the ranking table only.

        """
        if is_notebook():  # pragma: no cover
            logger.info(
                f"Training time was ({(time.time() - self.train_start_time):.2f} seconds.)"
            )

            if len(self.est.tuning_trials_) == 0 or len(self.est.train_shape_) == 0:
                logger.error(
                    "Unfortunately, there were no trials found, so we cannot visualize it."
                )
                return

            info = [
                ["Training Dataset size", self.X_train.shape],
                [
                    "Validation Dataset size",
                    self.X_valid.shape if self.X_valid is not None else None,
                ],
                ["CV", self.est.num_cv_folds_],
                ["Target variable", self.y_train.name],
                ["Optimization Metric", self.est.inferred_score_metric],
                ["Initial number of Features", self.est.train_shape_[1]],
                ["Selected number of Features", len(self.est.selected_features_names_)],
                ["Selected Features", self.est.selected_features_names_],
                ["Selected Algorithm", self.est.selected_model_],
                [
                    "End-to-end Elapsed Time (seconds)",
                    self.train_end_time - self.train_start_time,
                ],
                ["Selected Hyperparameters", self.est.selected_model_params_],
                ["Mean Validation Score", self.est.tuning_trials_[0][3]],
                ["AutoML n_jobs", self.est.n_jobs_],
                ["AutoML version", self.automl.__version__],
                ["Python version", sys.version],
            ]
            info_df = pd.DataFrame(info)

            # Remove the selected model and its params from the trials since it already shows up in the summary table
            all_trials_ = (
                self.est.model_selection_trials_
                + self.est.adaptive_sampling_trials_
                + self.est.feature_selection_trials_
                + self.est.tuning_trials_[1:]
            )
            col_names = [
                "Algorithm",
                "#Samples",
                "#Features",
                "Mean Validation Score",
                "Hyperparameters",
                "All Validation Scores",
                "CPU Time",
                "Memory Usage",
            ]
            if ranking_table_only:
                dropped_cols = [
                    "#Samples",
                    "#Features",
                    "All Validation Scores",
                    "CPU Time",
                ]
            else:
                dropped_cols = "All Validation Scores"
            summary_df = pd.DataFrame(all_trials_, columns=col_names).drop(
                dropped_cols, axis=1
            )
            sorted_summary_df = summary_df.sort_values(sort_column, ascending=False)
            # Add a rank column at the front
            sorted_summary_df.insert(
                0, "Rank based on Performance", np.arange(2, len(sorted_summary_df) + 2)
            )

            from IPython.core.display import display, HTML

            with pd.option_context(
                "display.max_colwidth",
                1000,
                "display.width",
                None,
                "display.precision",
                4,
            ):
                display(HTML(info_df.to_html(index=False, header=False)))
                if max_rows is None:
                    display(HTML(sorted_summary_df.to_html(index=False)))
                else:
                    display(
                        HTML(sorted_summary_df.to_html(index=False, max_rows=max_rows))
                    )

    def train(self, **kwargs):
        """
        Train the Oracle AutoML Pipeline. This looks at the training data, and
        identifies the best set of features, the best algorithm and the best
        set of hyperparameters for this data. A model is then generated, trained
        on this data and returned.

        Parameters
        ----------
        score_metric : str, callable
            Score function (or loss function) with signature ``score_func(y, y_pred, **kwargs)`` or string specified as
            https://scikit-learn.org/stable/modules/model_evaluation.html#common-cases-predefined-values
        random_state : int
            Random seed used by AutoML
        model_list : list of str
            Models that will be evaluated by the Pipeline. Supported models:
            - Classification: AdaBoostClassifier, DecisionTreeClassifier,
            ExtraTreesClassifier, KNeighborsClassifier,
            LGBMClassifier, LinearSVC, LogisticRegression,
            RandomForestClassifier, SVC, XGBClassifier
            - Regression: AdaBoostRegressor, DecisionTreeRegressor,
            ExtraTreesRegressor, KNeighborsRegressor,
            LGBMRegressor, LinearSVR, LinearRegression, RandomForestRegressor,
            SVR, XGBRegressor
        time_budget : float, optional
            Time budget in seconds where 0 means no time budget constraint (best effort)
        min_features : int, float, list, optional (default: 1)
            Minimum number of features to keep. Acceptable values:
            - If int, 0 < min_features <= n_features
            - If float, 0 < min_features <= 1.0
            - If list, names of features to keep, for example ['a', 'b'] means keep features 'a' and 'b'

        Returns
        -------
        self : object
        """

        """Adding this part to give the correct error for situations when dataset > 10000 rows and user tries SVC or KNN"""
        if len(self.X_train) > 10000:
            if "model_list" in kwargs:
                bad_model_list = ["SVC", "KNeighborsClassifier"]
                for model in kwargs["model_list"]:
                    for item in bad_model_list:
                        if item in model:
                            raise ValueError(
                                "SVC, KNeighborsClassifier are disabled for datasets with > 10K samples"
                            )

        self.train_start_time = time.time()

        self.time_budget = kwargs.pop("time_budget", 0)  # 0 means unlimited

        self.col_types = kwargs.pop("col_types", None)

        self.est = self._decide_estimator(**kwargs)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.est.fit(
                self.X_train,
                self.y_train,
                X_valid=self.X_valid,
                y_valid=self.y_valid,
                time_budget=self.time_budget,
                col_types=self.col_types,
            )
        self.train_end_time = time.time()
        self.print_summary(max_rows=10)

    def print_trials(self, max_rows=None, sort_column="Mean Validation Score"):
        """
        Prints all trials executed by the Oracle AutoML Pipeline in the last train() call.

        Parameters
        ----------
        max_rows : int
            Number of trials to print. Pass in None to print all trials
        sort_column: string
            Column to sort results by. Must be one of ['Algorithm', '#Samples', '#Features', 'Mean Validation Score',
            'Hyperparameters', 'All Validation Scores', 'CPU Time']

        """
        self.est.print_trials(max_rows=max_rows, sort_column=sort_column)

    def _decide_estimator(self, **kwargs):
        """
        Decide arguments to the Oracle AutoML pipeline based on user provided
        arguments
        """
        est = None
        score_metric = None
        # Explicity define the default AutoML metrics
        if (
            self.ml_task_type == ml_task_types.BINARY_CLASSIFICATION
            or self.ml_task_type == ml_task_types.BINARY_TEXT_CLASSIFICATION
        ):
            test_model_list = ["LogisticRegression"]
        elif (
            self.ml_task_type == ml_task_types.MULTI_CLASS_CLASSIFICATION
            or self.ml_task_type == ml_task_types.MULTI_CLASS_TEXT_CLASSIFICATION
        ):
            test_model_list = ["LogisticRegression"]
        elif self.ml_task_type == ml_task_types.REGRESSION:
            test_model_list = ["LinearRegression"]
        else:
            raise ValueError("AutoML for (%s) is not supported" % self.ml_task_type)

        # Respect the user provided scoring metric if given
        if "score_metric" in kwargs:
            score_metric = kwargs.pop("score_metric")

        #
        # ***FOR TESTING PURPOSE ONLY***
        #
        # Ignore model_list for test mode
        if ads.test_mode:  # pragma: no cover
            if "model_list" in kwargs:
                _ = kwargs.pop("model_list")
            kwargs["model_list"] = test_model_list

        if (
            self.ml_task_type == ml_task_types.BINARY_CLASSIFICATION
            or self.ml_task_type == ml_task_types.MULTI_CLASS_CLASSIFICATION
        ):
            est = self.automl.Pipeline(
                task="classification", score_metric=score_metric, **kwargs
            )
        elif (
            self.ml_task_type == ml_task_types.BINARY_TEXT_CLASSIFICATION
            or self.ml_task_type == ml_task_types.MULTI_CLASS_TEXT_CLASSIFICATION
        ):
            est = self.automl.Pipeline(
                task="classification", score_metric=score_metric, **kwargs
            )
            if not self.col_types:
                if len(self.X_train.columns) == 1:
                    self.col_types = ["text"]
                elif len(self.X_train.columns) == 2:
                    self.col_types = ["text", "text"]
                else:
                    raise ValueError(
                        "We detected a text classification problem. Pass "
                        "in `col_types = [<type of column1>, <type of column2>, ...]`."
                        " Valid types are: ['categorical', 'numerical', 'text', 'datetime',"
                        " 'timedelta']."
                    )

        elif self.ml_task_type == ml_task_types.REGRESSION:
            est = self.automl.Pipeline(
                task="regression", score_metric=score_metric, **kwargs
            )
        else:
            raise ValueError("AutoML for (%s) is not supported" % self.ml_task_type)
        return est

    def selected_score_label(self):
        """
        Return the name of score_metric used in train.

        """
        score_label = self.est.score_metric
        if score_label is None:
            score_label = self.est.inferred_score_metric
        return score_label

    @runtime_dependency(module="scipy", install_from=OptionalDependency.VIZ)
    def visualize_algorithm_selection_trials(self, ylabel=None):
        """
        Plot the scores predicted by Algorithm Selection for each algorithm. The
        horizontal line shows the average score across all algorithms. Algorithms
        below the line are colored turquoise, whereas those with a score higher
        than the mean are colored teal. The orange bar shows the algorithm with
        the highest predicted score. The error bar is +/- one standard error.

        Parameters
        ----------
        ylabel : str,
            Label for the y-axis. Defaults to the scoring metric.
        """
        if ylabel is None:
            ylabel = self.selected_score_label().capitalize()
        trials = self.est.model_selection_trials_
        if not len(trials):
            _log_visualize_no_trials("algorithm selection")
            return
        fig, ax = plt.subplots(1, figsize=(6, 3))
        colors = []
        y_error = []
        mean_scores, models, cvscores = [], [], []
        for (
            algorithm,
            samples,
            features,
            mean_score,
            hyperparameters,
            all_scores,
            runtime,
            x,
        ) in trials:
            mean_scores.append(mean_score)
            models.append(algorithm)
            cvscores.append(all_scores)
        mean_scores_ser = pd.Series(mean_scores, index=models).sort_values(
            ascending=False
        )
        scores_ser = pd.Series(cvscores, index=models)
        ax.set_title("Algorithm Selection Trials")
        ax.set_ylabel(ylabel)
        ax.set_xlabel("Algorithm")
        for f in mean_scores_ser.keys():
            se = scipy.stats.sem(scores_ser[f], ddof=1)
            y_error.append(se)
            if f == "{}_AS".format(self.est.selected_model_):
                colors.append("orange")
            elif mean_scores_ser[f] >= mean_scores_ser.mean():
                colors.append("teal")
            else:
                colors.append("turquoise")
        mean_scores_ser.plot.bar(ax=ax, color=colors, edgecolor="black", zorder=1)
        ax.errorbar(
            x=mean_scores_ser.index.values,
            y=mean_scores_ser.values,
            yerr=y_error,
            fmt="none",
            capsize=4,
            color="black",
            zorder=0,
        )
        ax.axhline(y=mean_scores_ser.mean(), color="black", linewidth=0.5)
        ax.autoscale_view()
        plt.show()

    def visualize_adaptive_sampling_trials(self):
        """
        Visualize the trials for Adaptive Sampling.
        """
        trials = self.est.adaptive_sampling_trials_
        if len(trials) == 0:
            _log_visualize_no_trials("adaptive sampling")
            return
        fig, ax = plt.subplots(1, figsize=(6, 3))
        ax.set_title("Adaptive Sampling ({})".format(trials[0][0]))
        ax.set_xlabel("Dataset sample size")
        ax.set_ylabel(r"Predicted model score")
        scores = [
            mean_score
            for (
                algorithm,
                samples,
                features,
                mean_score,
                hyperparameters,
                all_scores,
                runtime,
                x,
            ) in trials
        ]
        n_samples = [
            samples
            for (
                algorithm,
                samples,
                features,
                mean_score,
                hyperparameters,
                all_scores,
                runtime,
                x,
            ) in trials
        ]
        y_margin = 0.10 * (max(scores) - min(scores))
        ax.grid(color="g", linestyle="-", linewidth=0.1)
        ax.set_ylim(min(scores) - y_margin, max(scores) + y_margin)
        ax.plot(n_samples, scores, "k:", marker="s", color="teal", markersize=3)
        plt.show()

    def visualize_feature_selection_trials(self, ylabel=None):
        """
        Visualize the feature selection trials taken to arrive at optimal set of
        features. The orange line shows the optimal number of features chosen
        by Feature Selection.

        Parameters
        ----------
        ylabel : str,
            Label for the y-axis. Defaults to the scoring metric.
        """
        if ylabel is None:
            ylabel = self.selected_score_label().capitalize()
        trials = self.est.feature_selection_trials_
        if len(trials) == 0:
            _log_visualize_no_trials("feature selection")
            return
        fig, ax = plt.subplots(1, figsize=(6, 3))
        ax.set_title("Feature Selection Trials")
        ax.set_xlabel("Number of Features")
        ax.set_ylabel(ylabel)
        scores = [
            mean_score
            for (
                algorithm,
                samples,
                features,
                mean_score,
                hyperparameters,
                all_scores,
                runtime,
                x,
            ) in trials
        ]
        n_features = [
            features
            for (
                algorithm,
                samples,
                features,
                mean_score,
                hyperparameters,
                all_scores,
                runtime,
                x,
            ) in trials
        ]
        y_margin = 0.10 * (max(scores) - min(scores))
        ax.grid(color="g", linestyle="-", linewidth=0.1)
        ax.set_ylim(min(scores) - y_margin, max(scores) + y_margin)
        ax.plot(n_features, scores, "k:", marker="s", color="teal", markersize=3)
        ax.axvline(
            x=len(self.est.selected_features_names_), color="orange", linewidth=2.0
        )
        plt.show()

    def visualize_tuning_trials(self, ylabel=None):
        """
        Visualize (plot) the hyperparamter tuning trials taken to arrive at the optimal
        hyper parameters. Each trial in the plot represents a particular
        hyperparamter combination.

        Parameters
        ----------
        ylabel : str,
            Label for the y-axis. Defaults to the scoring metric.
        """
        if ylabel is None:
            ylabel = self.selected_score_label().capitalize()
        # scores in trials are sorted decreasingly.
        # reversed(trails) : let the scores sort in increasing order from left to right.
        scores = [
            mean_score
            for (
                algorithm,
                samples,
                features,
                mean_score,
                hyperparameters,
                all_scores,
                runtime,
                x,
            ) in reversed(self.est.tuning_trials_)
            if mean_score and not np.isnan(mean_score)
        ]
        if not len(scores) > 1:
            raise RuntimeError("Insufficient tuning trials.")
        else:
            fig, ax = plt.subplots(1, figsize=(6, 3))
            ax.set_title("Hyperparameter Tuning Trials")
            ax.set_xlabel("Iteration $n$")
            ax.set_ylabel(ylabel)
            y_margin = 0.10 * (max(scores) - min(scores))
            ax.grid(color="g", linestyle="-", linewidth=0.1)
            ax.set_ylim(min(scores) - y_margin, max(scores) + y_margin)
            ax.plot(
                range(1, len(scores) + 1),
                scores,
                "k:",
                marker="s",
                color="teal",
                markersize=3,
            )
            plt.show()


class AutoMLPreprocessingTransformer(object):  # pragma: no cover
    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self, msg):
        self.msg = msg

    def fit(self, X):
        """
        Fits the preprocessing Transformer

        Parameters
        ----------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be predicted on

        Returns
        -------
        Self: Estimator
            The fitted estimator
        """
        return self

    def transform(self, X):
        """
        Runs the preprocessing transform function and returns the result

        Parameters
        ---------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be transformed

        Returns
        -------
        X: Dataframe or list-like
            The transformed Dataframe.
        """
        return X

    def _log_visualize_no_trials(target):
        logger.error(
            f"There are no trials. Therefore, the {target} cannot be visualized."
        )

    def __repr__(self):
        return self.msg


class AutoMLFeatureSelection(object):  # pragma: no cover
    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self, msg):
        self.msg = msg

    def fit(self, X):
        """
        Fits the baseline estimator

        Parameters
        ----------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be predicted on

        Returns
        -------
        Self: Estimator
            The fitted estimator
        """
        return self

    def transform(self, X):
        """
        Runs the Baselines transform function and returns the result

        Parameters
        ---------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be transformed

        Returns
        -------
        X: Dataframe or list-like
            The transformed Dataframe.
        """
        return X

    def __repr__(self):
        return self.msg



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import contextlib
import logging
import os
import shutil
import tempfile
import zipfile
from io import DEFAULT_BUFFER_SIZE
from urllib import request
from urllib.parse import urlparse

import fsspec
from ads.common.auth import default_signer

logger = logging.getLogger(__name__)


class Artifact:
    """Represents a OCI Data Science Job artifact.
    The Artifact class is designed to add an additional processing step on runtime/source code.
    before uploading it as data science job artifact.

    A sub-class should implement the build() method to do the additional processing.
    A sub-class is designed to be used with context manager so that the temporary files are cleaned up properly.

    For example, the NotebookArtifact implements the build() method to convert the notebook to python script.
    with NotebookArtifact(runtime) as artifact:
        # The build() method will be called when entering the context manager
        # The final artifact for the job will be stored in artifact.path
        upload_artifact(artifact.path)
        # Files are cleaned up when exit or if there is an exception.

    """

    CONST_DRIVER_UTILS = "driver_utils.py"

    def __init__(self, source, runtime=None) -> None:
        # Get the full path of source file if it is local file.
        if not urlparse(source).scheme:
            self.source = os.path.abspath(os.path.expanduser(source))
        else:
            self.source = source
        self.path = None
        self.temp_dir = None
        self.runtime = runtime

    def __str__(self) -> str:
        if self.path:
            return self.path
        return self.source

    def __enter__(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        self.build()
        return self

    def __exit__(self, *exc):
        if self.temp_dir:
            self.temp_dir.cleanup()

    @staticmethod
    def _write_file(fp, to_path):
        """Reads a file from a file-like object and write it to specific local path.

        Parameters
        ----------
        fp : file-like object
            The source of the file.
        to_path : path-like object
            Local destination path.
        """
        with open(to_path, "wb") as out_file:
            block_size = DEFAULT_BUFFER_SIZE * 8
            while True:
                block = fp.read(block_size)
                if not block:
                    break
                out_file.write(block)

    @staticmethod
    def _download_from_web(url, to_path):
        """Downloads a single file from http/https/ftp.

        Parameters
        ----------
        url : str
            The URL of the source file.
        to_path : path-like object
            Local destination path.
        """
        url_response = request.urlopen(url)
        with contextlib.closing(url_response) as fp:
            logger.debug("Downloading from %s", url)
            Artifact._write_file(fp, to_path)

    @staticmethod
    def copy_from_uri(uri, to_path, unpack=False, **storage_options):
        """Copy file(s) to local path

        Parameters
        ----------
        uri : URI of the file
            The URI of the source file or directory, which can be local path of OCI object storage URI.
        to_path : path-like object
            The local destination path.
            If this is a directory, the source file/directory will be placed under it.
        unpack : bool
            Indicate if zip or tar.gz file specified by the uri should be unpacked.
            This option has no effect on other files.
        storage_options :
            Storage options for fsspec.
            For OCI object storage, the default_signer from ads.common.auth will be used
            if storage option is not specified.

        Returns
        -------
        str or path-like object
            The actual path of file/directory at destination.
            For copying a single file and to_path is a filename, this will be the same as to_path.
            For copying a single file and to_path is a directory, this will be to_path + filename.
            For copying a directory, this will be to_path + directory name.
        """
        scheme = urlparse(uri).scheme
        # temp_dir is used only if the uri is zip/tar file
        with tempfile.TemporaryDirectory() as temp_dir:
            if unpack and (
                str(uri).endswith(".zip")
                or str(uri).endswith(".tar.gz")
                or str(uri).endswith(".tar")
                or str(uri).endswith(".tgz")
            ):
                unpack_path = to_path
                to_path = temp_dir
            else:
                unpack_path = None
            if scheme in ["http", "https", "ftp"]:
                if os.path.isdir(to_path):
                    to_path = os.path.join(to_path, os.path.basename(uri))
                Artifact._download_from_web(uri, to_path)
            else:
                if scheme == "oci" and not storage_options:
                    storage_options = default_signer()
                fs = fsspec.filesystem(scheme, **storage_options)
                if os.path.isdir(to_path):
                    to_path = os.path.join(
                        to_path, os.path.basename(str(uri).rstrip("/"))
                    )
                fs.get(uri, to_path, recursive=True)
            if unpack_path:
                shutil.unpack_archive(to_path, unpack_path)
                to_path = unpack_path
        return to_path

    def build(self):
        """Builds the runtime artifact in the temporary directory.
        Subclass should implement this method to:
        1. Process the runtime
        2. Set the self.path to the final artifact path

        Raises
        ------
        NotImplementedError
            When this method is not implemented in the subclass.
        """
        raise NotImplementedError()


class NotebookArtifact(Artifact):
    """Represents a NotebookRuntime job artifact"""

    CONST_DRIVER_SCRIPT = "driver_notebook.py"

    def __init__(self, source, runtime) -> None:
        super().__init__(source, runtime=runtime)

    def build(self):
        """Prepares job artifact for notebook runtime"""
        driver_script = os.path.join(
            os.path.dirname(__file__), "../../templates", self.CONST_DRIVER_SCRIPT
        )
        driver_utils = os.path.join(
            os.path.dirname(__file__), "../../templates", self.CONST_DRIVER_UTILS
        )
        notebook_path = os.path.join(self.temp_dir.name, os.path.basename(self.source))
        output_path = (
            os.path.join(
                self.temp_dir.name,
                str(os.path.basename(self.source)).split(".", maxsplit=1)[0],
            )
            + ".zip"
        )
        self.copy_from_uri(self.source, notebook_path)
        with zipfile.ZipFile(output_path, "w") as zip_file:
            zip_file.write(notebook_path, os.path.basename(notebook_path))
            zip_file.write(driver_script, os.path.basename(driver_script))
            zip_file.write(driver_utils, os.path.basename(driver_utils))
        self.path = output_path


class ScriptArtifact(Artifact):
    """Represents a ScriptRuntime job artifact"""

    def build(self):
        """Prepares job artifact for script runtime.
        If the source is a file, it will be returned as is.
        If the source is a directory, it will be compressed as a zip file.
        """
        source = self.copy_from_uri(self.source, self.temp_dir.name)
        # Zip the artifact if it is a directory
        if os.path.isdir(source):
            basename = os.path.basename(str(source).rstrip("/"))
            source = str(source).rstrip("/")
            # Runtime must have entrypoint if the source is a directory
            if self.runtime and not self.runtime.entrypoint:
                raise ValueError(
                    "Please specify entrypoint when script source is a directory."
                )
            output = os.path.join(self.temp_dir.name, basename)
            shutil.make_archive(
                output, "zip", os.path.dirname(source), base_dir=basename
            )
            self.path = output + ".zip"
            return
        # Otherwise, use the artifact directly
        self.path = source


class PythonArtifact(Artifact):
    """Represents a PythonRuntime job artifact"""

    CONST_DRIVER_SCRIPT = "driver_python.py"
    # Temp archive dir storing the zip/tar file from the user
    ARCHIVE_DIR = "archive"
    # The directory to store user code
    # This directory must match the USER_CODE_DIR in driver_python.py
    USER_CODE_DIR = "code"

    def build(self):
        """Prepares job artifact for PythonRuntime."""
        basename = os.path.basename(str(self.source).rstrip("/")).split(".", 1)[0]
        artifact_dir = os.path.join(self.temp_dir.name, basename)
        code_dir = os.path.join(artifact_dir, self.USER_CODE_DIR)
        os.makedirs(artifact_dir, exist_ok=True)

        # Copy the driver script
        for filename in [
            self.CONST_DRIVER_UTILS,
            self.CONST_DRIVER_SCRIPT,
            NotebookArtifact.CONST_DRIVER_SCRIPT,
        ]:
            file_path = os.path.join(
                os.path.dirname(__file__), "../../templates", filename
            )

            shutil.copy(file_path, os.path.join(artifact_dir, filename))

        # Copy user code
        os.makedirs(code_dir, exist_ok=True)
        Artifact.copy_from_uri(self.source, code_dir, unpack=True)

        # Check if entrypoint is valid
        # If the user code is a directory,
        # user should specify the entrypoint with the name of the top level directory.
        # For example, if the user code is in "path/to/dir"
        # The entrypoint should be something like dir/
        if self.runtime and self.runtime.entrypoint:
            if not os.path.exists(
                os.path.join(
                    code_dir, self.runtime.working_dir, self.runtime.entrypoint
                )
            ):
                # The specific entrypoint does not exist.
                # Check if user forgot to specify the top level directory.
                possible_entrypoint = os.path.join(
                    code_dir,
                    self.runtime.working_dir,
                    basename,
                    self.runtime.entrypoint,
                )
                err_message = (
                    f"Invalid entrypoint. {self.runtime.entrypoint} does not exist."
                )
                if os.path.exists(possible_entrypoint):
                    suggested_entrypoint = os.path.join(basename, self.runtime.entrypoint)
                    err_message += f" Do you mean {suggested_entrypoint}?"
                logger.warning(err_message)

        # Zip the job artifact
        output = os.path.join(self.temp_dir.name, basename)
        shutil.make_archive(output, "zip", artifact_dir, base_dir="./")
        self.path = output + ".zip"


class GitPythonArtifact(Artifact):

    CONST_DRIVER_SCRIPT = "driver_oci.py"

    def __init__(self) -> None:
        super().__init__("", runtime=None)

    def build(self):
        """Prepares job artifact for GitPythonRuntime."""
        artifact_dir = os.path.join(self.temp_dir.name, "ads_driver")
        os.makedirs(artifact_dir, exist_ok=True)
        for filename in [
            self.CONST_DRIVER_UTILS,
            self.CONST_DRIVER_SCRIPT,
            NotebookArtifact.CONST_DRIVER_SCRIPT,
        ]:
            file_path = os.path.join(
                os.path.dirname(__file__), "../../templates", filename
            )

            shutil.copy(file_path, os.path.join(artifact_dir, filename))

        # Zip the job artifact
        shutil.make_archive(artifact_dir, "zip", artifact_dir, base_dir="./")
        self.path = artifact_dir + ".zip"



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Dict, List

from ads.pipeline.visualizer.base import (
    COMPLETE_STATUS,
    STATUS_TEXT_MAP,
    PipelineRenderer,
    RendererItem,
    RendererItemStatus,
    StepStatus,
    StepKind,
)
from tabulate import tabulate


class PipelineTextRenderer(PipelineRenderer):
    def render(
        self,
        steps: List[RendererItem],
        deps: Dict[str, List[RendererItem]],
        step_status: Dict[str, RendererItemStatus] = None,
        **kwargs,
    ):
        """
        Render pipeline step status in text.

        Parameters
        ----------
        steps: List[RendererItem]
            A list of RendererItem objects.
        deps: Dict[str, List[RendererItem]]
            A dictionary mapping the key of a RendererItem to a list of
            RendererItem that this step depends on.
        step_status: Dict[str, RendererItemStatus], defaults to None.
            A dictionary mapping the key of a RendererItem to its current
            status.

        Returns
        -------
        None
        """
        print(
            tabulate(
                self._render(steps=steps, deps=deps, step_status=step_status, **kwargs),
                headers="keys",
            )
        )

    def _render(
        self,
        steps: List[RendererItem],
        deps: Dict[str, List[str]],
        step_status: Dict[str, RendererItemStatus],
        **kwargs,
    ) -> List[Dict[str, str]]:
        result = []
        steps = steps or []
        deps = deps or {}
        step_status = step_status or {}

        for step in steps:
            status = ""
            if step.kind != StepKind.PIPELINE:
                if step_status:
                    status = STATUS_TEXT_MAP[step_status[step.key].lifecycle_state]
                    if status.lower() == StepStatus.WAITING.lower():
                        unfinished_depended_steps = [
                            substep.name
                            for substep in deps.get(step.key, [])
                            if step_status[substep].lifecycle_state
                            not in COMPLETE_STATUS
                        ]
                        if unfinished_depended_steps == 0:
                            continue
                        status = f"{status} [{', '.join(unfinished_depended_steps)}]"

                result.append({"Step": f"{step.name}: ", "Status": status})

        return result



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import abstractmethod
from typing import List, Union

from ads.common import auth as authutil
from ads.common import oci_client
from ads.data_labeling.boundingbox import BoundingBoxItem
from ads.data_labeling.constants import AnnotationType, Formats
from ads.data_labeling.interface.parser import Parser
from ads.data_labeling.ner import NERItem
from ads.data_labeling.reader.dls_record_reader import OCIRecordSummary
from ads.data_labeling.record import Record
from oci.data_labeling_service_dataplane.models import (
    AnnotationSummary as OCIAnnotationSummary,
)
from oci.data_labeling_service_dataplane.models.annotation import (
    Annotation as OCIAnnotation,
)
from oci.exceptions import ServiceError


class ReadAnnotationError(Exception):
    def __init__(self, id: str):
        super().__init__(f"Error occurred in attempt to read annotation {id}.")


class AnnotationNotFoundError(Exception):
    def __init__(self, id: str):
        super().__init__(f"{id} not found.")


class DLSRecordParser(Parser):
    """DLSRecordParser class which parses the labels from the record."""

    def __init__(
        self,
        dataset_source_path: str,
        auth: dict = None,
        format: str = None,
        categories: List[str] = None,
    ) -> "DLSRecordParser":
        """Initiates a DLSRecordParser instance.

        Parameters
        ----------
        dataset_source_path: str
            Dataset source path.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        format: (str, optional). Defaults to None.
            Output format of annotations.
        categories: (List[str], optional). Defaults to None.
            The list of object categories in proper order for model training.
            Example: ['cat','dog','horse']

        Returns
        -------
        DLSRecordParser
            DLSRecordParser instance.
        """
        auth = auth or authutil.default_signer()
        self.dataset_source_path = dataset_source_path
        self.format = format
        self.categories = categories
        self.dls_dp_client = oci_client.OCIClientFactory(**auth).data_labeling_dp

    def _get_annotation_details(
        self, oci_annotation_summary: List[OCIAnnotationSummary]
    ) -> List[OCIAnnotation]:
        """Get list of annotation details.

        Parameters
        ----------
        oci_annotation_summary: List[OCIAnnotationSummary]
            The list of OCIAnnotationSummary objects.

        Returns
        -------
        List[OCIAnnotation]
            List of OCI annotations.
        """
        result = []
        if oci_annotation_summary:
            for annotation in oci_annotation_summary:
                try:
                    annotation_details = self.dls_dp_client.get_annotation(
                        annotation.id
                    ).data
                except ServiceError as service_err:
                    if service_err.status == 404:
                        raise AnnotationNotFoundError(annotation.id)
                    raise ReadAnnotationError(annotation.id)
                result.append(annotation_details)
        return result

    def parse(self, oci_record_summary: OCIRecordSummary) -> Record:
        """Extracts the annotations.
        Constructs and returns a Record instance which contains the file path and the labels.

        Parameters
        ----------
        oci_record_summary: OCIRecordSummary
            The summary information about the record.

        Returns
        -------
        Record
            Record instance which contains the file path and the annotations.
        """
        return Record(
            path=self.dataset_source_path + oci_record_summary.record.name,
            annotation=self._extract_annotations(
                self._get_annotation_details(oci_record_summary.annotation),
                format=self.format,
                categories=self.categories,
            ),
        )

    @abstractmethod
    def _extract_annotations(
        self, oci_annotation: List[OCIAnnotation], **kwargs
    ) -> Union[str, List[str], List[BoundingBoxItem], List[NERItem]]:
        """Extracts annotations from the record content. Each Parser class
        needs to implement this function.

        Parameters
        ----------
        oci_annotation: List[OCIAnnotation]
            List of OCI annotations.
        kwargs: Dict
            format: str
                Output format of annotations. Can be "spacy" or "yolo".

        Returns
        -------
        Union[str, List[str], List[BoundingBoxItem], List[NERItem]]
            Label(s).
        """
        pass


class DLSSingleLabelRecordParser(DLSRecordParser):
    """SingleLabelRecordParser class which parses the label of Single label data."""

    def _extract_annotations(
        self, oci_annotation: List[OCIAnnotation], **kwargs
    ) -> Union[str, None]:
        """Extract the labels of the single label annotation class.

        Parameters
        ----------
        oci_annotation: List[OCIAnnotation]
            List of OCI annotations.

        Returns
        -------
        Union[str, None]
            A label or None for the unlabeled record.
        """
        if oci_annotation:
            return oci_annotation[0].entities[0].labels[0].label
        else:
            return None


class DLSMultiLabelRecordParser(DLSRecordParser):
    """MultiLabelRecordParser class which parses the label of Multiple label data."""

    def _extract_annotations(
        self, oci_annotation: List[OCIAnnotation], **kwargs
    ) -> Union[List[str], None]:
        """Extract labels of the Multi label annotation class.

        Parameters
        ----------
        oci_annotation: List[OCIAnnotation]
            List of OCI annotations.

        Returns
        -------
        Union[List[str], None]
            List of labels or None for the unlabeled record.
        """
        if oci_annotation:
            return [
                annotation_item.label
                for annotation_item in oci_annotation[0].entities[0].labels
            ]
        else:
            return None


class DLSNERRecordParser(DLSRecordParser):
    """NERRecordParser class which parses the label of NER label data."""

    def _extract_annotations(
        self, oci_annotation: List[OCIAnnotation], **kwargs
    ) -> Union[List[NERItem], None]:
        """Extracts the labels of the NER annotation class.

        Parameters
        ----------
        oci_annotation: List[OCIAnnotation]
            List of OCI annotations.
        kwargs: Dict
            format: str
                Output format of annotations. Can be "spacy" or None.
                When None, it outputs List[NERItem]. When "spacy", it
                outputs List[Tuple].

        Returns
        -------
        Union[List[NERItem], None]
            The list of NERItem objects.
        """
        items = []
        format = kwargs.get("format", None)
        if oci_annotation:
            for e in oci_annotation[0].entities:
                label = e.labels[0].label
                offset = e.text_span.offset
                length = e.text_span.length
                item = NERItem(label=label, offset=int(offset), length=int(length))
                if (
                    format
                    and isinstance(format, str)
                    and format.lower() == Formats.SPACY
                ):
                    item = item.to_spacy()
                items.append(item)
        return items


class DLSBoundingBoxRecordParser(DLSRecordParser):
    """BoundingBoxRecordParser class which parses the label of BoundingBox label data."""

    def _extract_annotations(
        self, oci_annotation: List[OCIAnnotation], **kwargs
    ) -> Union[List[BoundingBoxItem], None]:
        """Extracts the labels of the Object Detection annotation class.

        Parameters
        ----------
        oci_annotation: List[OCIAnnotation]
            List of OCI annotations.
        kwargs: Dict
            format: str
                Output format of annotations. Can be None or "yolo".
                When None, it outputs List[BoundingBoxItem]. When "yolo", it
                outputs List[List[Tuple]].
            categories: (List[str], optional).
                 The list of object categories in proper order for model training.
                 Only used when bounding box annotations are in YOLO format.

        Returns
        -------
        Union[List[BoundingBoxItem], None]
            The list of BoundingBoxItem objects.
        """
        format = kwargs.get("format", None)
        categories = kwargs.get("categories", None)

        items = []
        if oci_annotation:
            for e in oci_annotation[0].entities:
                labels = [l.label for l in e.labels]
                coords = e.bounding_polygon.normalized_vertices
                top_left = (float(coords[0].x), float(coords[0].y))
                bottom_left = (float(coords[1].x), float(coords[1].y))
                bottom_right = (float(coords[2].x), float(coords[2].y))
                top_right = (float(coords[3].x), float(coords[3].y))
                item = BoundingBoxItem(
                    labels=labels,
                    bottom_left=bottom_left,
                    top_left=top_left,
                    top_right=top_right,
                    bottom_right=bottom_right,
                )
                if (
                    format
                    and isinstance(format, str)
                    and format.lower() == Formats.YOLO
                ):
                    item = item.to_yolo(categories=categories)
                items.append(item)

        return items


class DLSRecordParserFactory:
    """DLSRecordParserFactory class which contains a list of registered parsers
    and allows to register new DLSRecordParsers.

    Current parsers include:
        * DLSSingleLabelRecordParser
        * DLSMultiLabelRecordParser
        * DLSNERRecordParser
        * DLSBoundingBoxRecordParser
    """

    _parsers = {
        AnnotationType.SINGLE_LABEL: DLSSingleLabelRecordParser,
        AnnotationType.MULTI_LABEL: DLSMultiLabelRecordParser,
        AnnotationType.ENTITY_EXTRACTION: DLSNERRecordParser,
        AnnotationType.BOUNDING_BOX: DLSBoundingBoxRecordParser,
    }

    @staticmethod
    def parser(
        annotation_type: str,
        dataset_source_path: str,
        format: str = None,
        categories: List[str] = None,
        auth: dict = None,
    ) -> "DLSRecordParser":
        """Gets the parser based on the annotation_type.

        Parameters
        ----------
        annotation_type: str
            Annotation type which can be SINGLE_LABEL, MULTI_LABEL, ENTITY_EXTRACTION
            and BOUNDING_BOX.
        dataset_source_path: str
            Dataset source path.
        format: (str, optional). Defaults to None.
            Output format of annotations. Can be None, "spacy" for dataset
            Entity Extraction type or "yolo" for Object Detection type.
            When None, it outputs List[NERItem] or List[BoundingBoxItem].
            When "spacy", it outputs List[Tuple].
            When "yolo", it outputs List[List[Tuple]].
        categories: (List[str], optional). Defaults to None.
            The list of object categories in proper order for model training.
            Example: ['cat','dog','horse']
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        DLSRecordParser
            DLSRecordParser corresponding to the annotation type.

        Raises
        ------
        ValueError
            If annotation_type is not supported.
        """

        if not annotation_type in DLSRecordParserFactory._parsers:
            raise ValueError(
                f"The {annotation_type} is not supported. Choose from "
                f"`{AnnotationType.SINGLE_LABEL}`, `{AnnotationType.MULTI_LABEL}`, "
                f"`{AnnotationType.ENTITY_EXTRACTION}` and `{AnnotationType.BOUNDING_BOX}`."
            )

        return DLSRecordParserFactory._parsers[annotation_type](
            dataset_source_path=dataset_source_path,
            format=format,
            categories=categories,
            auth=auth or authutil.default_signer(),
        )

    @classmethod
    def register(cls, annotation_type: str, parser: DLSRecordParser) -> None:
        """Registers a new parser.

        Parameters
        ----------
        annotation_type: str
            Annotation type which can be SINGLE_LABEL, MULTI_LABEL, ENTITY_EXTRACTION
            and BOUNDING_BOX.
        parser: DLSRecordParser
            A new Parser class to be registered.

        Returns
        -------
        None
            Nothing.
        """
        cls._parsers[annotation_type] = parser



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Any, Dict, Generator
from ads.data_labeling.reader.jsonl_reader import JsonlReader


class ExportRecordReader(JsonlReader):
    """The ExportRecordReader class to read labeled dataset records from the export.

    Methods
    -------
    read(self) -> Generator[Dict, Any, Any]
        Reads labeled dataset records.
    """

    def __init__(
        self,
        path: str,
        auth: Dict = None,
        encoding="utf-8",
        includes_metadata: bool = False,
    ) -> "ExportRecordReader":
        """Initiates an ExportRecordReader instance.

        Parameters
        ----------
        path: str
            object storage path or local path for a file.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        encoding : (str, optional). Defaults to 'utf-8'.
            Encoding of files. Only used for "TEXT" dataset.
        includes_metadata: (bool, optional). Defaults to False.
            Determines whether the export file includes metadata or not.

        Examples
        --------
        >>> from ads.data_labeling.reader.export_record_reader import ExportRecordReader
        >>> path = "your/path/to/jsonl/file.jsonl"
        >>> from ads.common import auth as authutil
        >>> reader = ExportRecordReader(path=path, auth=authutil.api_keys(), encoding="utf-8")
        >>> next(reader.read())
        """
        super().__init__(path=path, auth=auth, encoding=encoding)
        self._includes_metadata = includes_metadata

    def read(self) -> Generator[Dict, Any, Any]:
        """Reads labeled dataset records.

        Returns
        -------
        Generator[Dict, Any, Any]
            The labeled dataset records.
        """
        skip = 1 if self._includes_metadata else None
        return super().read(skip=skip)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import importlib
import json
import os
import pickle
import shutil
import tempfile
import uuid

from ads.common import oci_client
from ads.common import auth as authutil

import sklearn
from ads.hpo.ads_search_space import model_list
from ads.hpo.distributions import decode, encode
from ads.hpo.utils import _extract_uri

from ads.common import logger
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class NotPickableError(Exception):
    def __init__(self, message):
        self.message = message


class UploadTunerArtifact:
    def __init__(self, tuner, file_uri, metadata, auth=None):
        self.tuner = tuner
        self.file_uri = file_uri
        self.metadata = metadata
        self.oci_client = None
        self.tuner_args = None
        self.zip_dir = None
        self.auth = auth if auth else authutil.default_signer()

    def upload(self, script_dict):
        tuner_zip = self.prepare_tuner_artifact(script_dict)
        self.upload_to_cloud(tuner_zip)

    def prepare_tuner_artifact(self, script_dict):
        """
        zip and save all the tuner files

        Args:
            script_dict (dict): dict which contains the script names and path
        """
        tuner_path = "tuner_artifacts" + str(uuid.uuid4()) + ".zip"
        tuner_path_name = os.path.splitext(tuner_path)[0]

        with tempfile.TemporaryDirectory() as temp_dir:
            # make the zip dir is not exist
            self.zip_dir = os.path.join(temp_dir, tuner_path_name)
            logger.debug(f"create the temp dir: {self.zip_dir}")
            if not os.path.exists(self.zip_dir):
                os.makedirs(self.zip_dir)
            logger.debug(
                f"start serializing the tuner argument and save at tuner.json ..."
            )
            self.tuner_args = self.json_serialize_tuner_args(script_dict)
            logger.debug(f"start saving tuner.json at {tuner_path}...")
            self._save_tuner_args_as_json()
            logger.debug(f"start saving the scipts at {tuner_path} if there are any...")
            self._save_script(script_dict)
            if "sqlite:///" in self.tuner.storage:
                if "sqlite:///:memory:" in self.tuner.storage:
                    raise NotImplementedError(
                        "Cannot save in-memory sqlite database to the object storage. Store it on block storage instead."
                    )
                shutil.copy(
                    os.path.join(self.tuner.storage.replace("sqlite:///", "")),
                    self.zip_dir,
                )
            else:
                raise NotImplementedError(
                    "Only sqlite file is supported to be saved to the object storage."
                )
            logger.debug(f"start archiving the files at {tuner_path}...")
            shutil.make_archive(
                tuner_path_name, "zip", base_dir=tuner_path_name, root_dir=temp_dir
            )
        return tuner_path

    def json_serialize_tuner_args(self, script_dict):
        """
        json serialize the tuner args
        """
        tuner_args = {}
        serialising_fields = [
            "model",
            "scoring",
            "strategy",
            "cv",
            "loglevel",
            "storage",
            "study_name",
            "__random_state",
            "load_if_exists",
            "n_jobs",
        ]
        for field in serialising_fields:
            if field == "strategy":
                tuner_args["strategy"] = encode(self.tuner.strategy)
            elif field == "model":
                tuner_args["model"] = self._serialize_model(script_dict)
            elif field == "scoring" and self.tuner.scoring is not None:
                tuner_args["scoring"] = self._serialize_scoring(script_dict)
            else:
                try:
                    if getattr(self.tuner, field) is not None:
                        tuner_args[field] = getattr(self.tuner, field)
                except:
                    logger.debug(
                        field.replace("__", "") + " is not serialized since it is None."
                    )
        return tuner_args

    def _serialize_model(self, script_dict):
        if "model" in script_dict and script_dict["model"] is not None:
            # save as script
            model = {"model_script": script_dict["model"]}
        elif self.tuner.model.__class__ in model_list:
            model = {
                "model_name": self.tuner.model.__class__.__name__,
                "model_input": self.tuner.model.get_params(),
            }
        else:
            # save as pickle
            try:
                model = {"pickled_model_path": "model.pkl"}
                with open(
                    os.path.join(self.zip_dir, model["pickled_model_path"]), "wb"
                ) as f:
                    pickle.dump(self.tuner.model, f)
            except:
                raise NotPickableError(
                    "Model is not pickable. Try to save the model in a script."
                )
        return model

    def _serialize_scoring(self, script_dict):
        if isinstance(self.tuner.scoring, str):
            scoring = {"scoring_name": self.tuner.scoring}
        else:
            assert (
                script_dict["scoring"] is not None
            ), "Need to save the customized scoring function \
                    in a script and pass the script name to <code/>script_dict={'scoring': script_name}<code>"
            scoring = {"scoring_script": script_dict["scoring"]}
        return scoring

    def _save_script(self, script_dict):
        for var, script_path in script_dict.items():
            if script_path is not None:
                # script_name = os.path.basename(script_path)
                #
                shutil.copy(script_path, self.zip_dir)

    def _save_tuner_args_as_json(self):
        try:
            with open(os.path.join(self.zip_dir, "tuner.json"), "w") as outfile:
                json.dump(self.tuner_args, outfile)
        except:
            raise RuntimeError(f"Failed to save .json at this path: {self.zip_dir}.")

    def upload_to_cloud(self, tuner_zip):
        bucketname, namespace, filename = _extract_uri(self.file_uri)
        if self.oci_client is None:
            self.oci_client = oci_client.OCIClientFactory(**self.auth).object_storage

        try:
            with open(tuner_zip, "rb") as zipfile:
                self.oci_client.put_object(
                    namespace,
                    bucketname,
                    filename,
                    zipfile,
                    opc_meta={"info": self.metadata},
                )
        finally:
            logger.debug(f"removing {tuner_zip}")
            os.remove(tuner_zip)


class DownloadTunerArtifact:
    """
    Download the tuner artifact from the cloud and deserialize the tuner args
    """

    def __init__(self, file_uri, target_file_path=None, auth=None):
        self.file_uri = file_uri
        self.target_file_path = (
            os.path.join("/tmp", "hpo_" + str(uuid.uuid4()) + ".zip")
            if target_file_path is None
            else target_file_path
        )
        self.oci_client = None
        self.tuner_args = None
        self.metadata = None
        self.auth = auth if auth else authutil.default_signer()

    def extract_tuner_args(self, delete_zip_file=False):
        """
        deserialize tuner argument from the zip file
        """
        self.download_from_cloud()
        file_path_dict = self._find_file_path()
        tuner_args = self._load_json(file_path_dict["json_path"])
        if tuner_args is not None:
            self._save_db_file(file_path_dict)
            transformed_args = self.deserialize_tuner_args(file_path_dict)
            if delete_zip_file:
                os.remove(self.target_file_path)
                shutil.rmtree(os.path.splitext(self.target_file_path)[0])
            return transformed_args, self.metadata
        else:
            raise RuntimeError("Tuner args are not found or loaded.")

    def download_from_cloud(self):
        """
        Download the artifact and unpack the arhchive at the target file path
        """
        if os.path.isdir(self.target_file_path):
            raise ValueError(
                "`target_file_path` should be a file, for example, '/home/datascience/myfile.zip', but given a dir."
            )
        if not self.target_file_path.endswith(".zip"):
            self.target_file_path += ".zip"
        bucketname, namespace, filename = _extract_uri(self.file_uri)
        if self.oci_client is None:
            self.oci_client = oci_client.OCIClientFactory(**self.auth).object_storage

        res = self.oci_client.get_object(namespace, bucketname, filename)
        self.metadata = res.data.headers["opc-meta-info"]

        if os.path.exists(self.target_file_path):
            raise FileExistsError(
                "This file name already exists, do you want to delete it or use a different file path?"
            )
        target_dir = os.path.dirname(self.target_file_path)
        if not os.path.exists(target_dir):
            os.path.makedirs(target_dir)
        with open(self.target_file_path, "wb") as dbfile:
            for chunk in res.data.iter_content(chunk_size=4096):
                dbfile.write(chunk)
        shutil.unpack_archive(
            self.target_file_path, os.path.splitext(self.target_file_path)[0]
        )

    def _find_file_path(self):
        file_path_dict = {}
        for subdir, dirs, files in os.walk(os.path.splitext(self.target_file_path)[0]):
            for file in files:
                file_path = os.path.join(subdir, file)
                if file.endswith(".json"):
                    file_path_dict["json_path"] = file_path
                if file.endswith(".db"):
                    file_path_dict["db_path"] = file_path
        return file_path_dict

    def _load_json(self, json_path):
        if os.path.exists(json_path):
            with open(json_path, "r") as outfile:
                self.tuner_args = json.load(outfile)
        else:
            raise FileNotFoundError(f"{json_path} does not exist.")
        return self.tuner_args

    def _save_db_file(self, file_path_dict):
        """
        save the database file to the same original path, if the path does not exist, create one.

        Args:
            file_path_dict (dict): dict which contains the path of different files
        """
        storage_dest_folder = os.path.dirname(
            self.tuner_args["storage"].replace("sqlite:///", "")
        )
        if not os.path.exists(storage_dest_folder):
            logger.debug(f"{self.storage_dest_folder} does not exist, creating one.")
            os.path.makedirs(storage_dest_folder)
        if os.path.exists(file_path_dict["db_path"]):
            shutil.copy(file_path_dict["db_path"], storage_dest_folder)
        else:
            raise FileNotFoundError(f'{file_path_dict["db_path"]} does not exist.')

    def deserialize_tuner_args(self, file_path_dict):
        """
        deserialize the tuner args

        Args:
            file_path_dict (dict): dict which contains the path of different files
        """
        self._does_model_exist()
        tuner_input = {}
        for field, value in self.tuner_args.items():
            if field == "model":
                tuner_input["model"] = self._deserialize_model(file_path_dict)
            elif field == "scoring":
                tuner_input["scoring"] = self._deserialize_scoring(file_path_dict)
            elif field == "strategy":
                tuner_input["strategy"] = decode(self.tuner_args["strategy"])
            elif field in ["loglevel", "storage", "study_name"]:
                tuner_input[field] = value
        return tuner_input

    def _does_model_exist(self):
        assert "model" in self.tuner_args, "`model` not found."

    def _deserialize_model(self, file_path_dict):
        model_args = self.tuner_args["model"]
        model_dir = os.path.dirname(file_path_dict["json_path"])
        if "model_name" in model_args:
            model = get_supported_model_mappings()[model_args["model_name"]](
                **model_args["model_input"]
            )
        elif "pickled_model_path" in model_args:
            model_path = os.path.join(model_dir, os.path.basename("model.pkl"))
            model = pickle.load(open(model_path, "rb"))
        elif "model_script" in model_args:
            model_path = os.path.join(
                model_dir, os.path.basename(model_args["model_script"])
            )
            model = self.load_model(model_path)
        else:
            raise RuntimeError("model cannot be found.")
        return model

    def _deserialize_scoring(self, file_path_dict):
        scoring_args = self.tuner_args["scoring"]
        if "scoring_name" in scoring_args:
            scoring = scoring_args["scoring_name"]
        elif "scoring_script" in scoring_args:
            scoring_path = os.path.join(
                os.path.dirname(file_path_dict["json_path"]),
                os.path.basename(scoring_args["scoring_script"]),
            )
            scoring = self.load_scoring(scoring_path)
        else:
            raise RuntimeError("scoring cannot be found.")
        return scoring

    @staticmethod
    def load_target_from_script(script_path):
        script_name = os.path.basename(script_path).replace(".py", "")
        spec = importlib.util.spec_from_file_location(
            script_name + "%s" % uuid.uuid4(), script_path
        )
        func = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(func)
        return func

    @staticmethod
    def load_model(script_path):
        spec = DownloadTunerArtifact.load_target_from_script(script_path)
        if hasattr(spec, "model"):
            model = spec.model
        else:
            raise Exception(
                "Need to explicitly define `model=your_model` in the model script."
            )
        return model

    @staticmethod
    def load_scoring(script_path):
        spec = DownloadTunerArtifact.load_target_from_script(script_path)
        if hasattr(spec, "scoring"):
            scoring = spec.scoring
        else:
            raise Exception(
                "Need to explicitly define `scoring=your_scoring_func` in the model script."
            )
        return scoring


@runtime_dependency(module="lightgbm", install_from=OptionalDependency.BOOSTED)
@runtime_dependency(module="xgboost", install_from=OptionalDependency.BOOSTED)
def get_supported_model_mappings():

    supported_model_mapping = {
        "Ridge": sklearn.linear_model._ridge.Ridge,
        "RidgeClassifier": sklearn.linear_model._ridge.RidgeClassifier,
        "Lasso": sklearn.linear_model._coordinate_descent.Lasso,
        "ElasticNet": sklearn.linear_model._coordinate_descent.ElasticNet,
        "LogisticRegression": sklearn.linear_model._logistic.LogisticRegression,
        "SVC": sklearn.svm._classes.SVC,
        "SVR": sklearn.svm._classes.SVR,
        "LinearSVC": sklearn.svm._classes.LinearSVC,
        "LinearSVR": sklearn.svm._classes.LinearSVR,
        "DecisionTreeClassifier": sklearn.tree._classes.DecisionTreeClassifier,
        "DecisionTreeRegressor": sklearn.tree._classes.DecisionTreeRegressor,
        "RandomForestClassifier": sklearn.ensemble._forest.RandomForestClassifier,
        "RandomForestRegressor": sklearn.ensemble._forest.RandomForestRegressor,
        "GradientBoostingClassifier": sklearn.ensemble._gb.GradientBoostingClassifier,
        "GradientBoostingRegressor": sklearn.ensemble._gb.GradientBoostingRegressor,
        "XGBClassifier": xgboost.sklearn.XGBClassifier,
        "XGBRegressor": xgboost.sklearn.XGBRegressor,
        "ExtraTreesClassifier": sklearn.ensemble._forest.ExtraTreesClassifier,
        "ExtraTreesRegressor": sklearn.ensemble._forest.ExtraTreesRegressor,
        "LGBMClassifier": lightgbm.sklearn.LGBMClassifier,
        "LGBMRegressor": lightgbm.sklearn.LGBMRegressor,
        "SGDClassifier": sklearn.linear_model._stochastic_gradient.SGDClassifier,
        "SGDRegressor": sklearn.linear_model._stochastic_gradient.SGDRegressor,
    }
    return supported_model_mapping
